/*->c.flex */


#include <stdafx.h>

#define BOOL int
#define TRUE 1
#define FALSE 0

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>

#include "os.h"
#include "err.h"
#include "wimp.h"


#include "wos.h"
#include "poll.h"
#include "etc.h"


#include "deb.h"
#include "osmem.h"


#include "flex.h"


#define FLEXEX



#define FIXEDMEM



#ifndef FIXEDMEM


/* There are two alternative implementations in this file. */


typedef struct
{
  flex_ptr    anchor;    /* *anchor should point back to here.    */
  int         size;      /* in bytes. Exact size of logical area. */

#ifdef FLEXEX

  flex_lockfn lock;      /* fn to call when block is moved        */
  int         handle;    /* user handle                           */
                         /* then the actual store follows.        */
#endif


} flex__rec;


/* From base upwards, it's divided into store blocks of
  a flex__rec
  the space
  align up to next word.
*/





static int roundup(int i)
{
 return(0xfffffffc & (i+3));
}


static flexblockstr flexbaseblock;





int flex_freemem(void)
{
 char * limit;
 int    free;

 Task_MemoryLimit(flexbaseblock.areano,(void**)&limit);
 free=(limit-flexbaseblock.flex__freep);
 free-=2*sizeof(flexblockstr);
 if(free<0) free=0;

 return(free);
}





static void flex__slot(int delta,flexblockstr * fb)
{
 if(delta)
 {
//  cprintf("slot oldtop %x newtop %x",fb->flex__lim,((char*)fb->flex__lim+delta));
  Task_ChangeMemoryAreaSize(fb->areano,fb->flex__lim,(void*)((char*)fb->flex__lim+delta),(void**)&fb->flex__lim);
//  cprintf("slot oldtop %x newtop %x",fb->flex__lim,fb->flex__lim);
 }
}


 /* Tries to get at least n more bytes, raising flex__lim and
  returning TRUE if it can. */

static os_error * flex__more(int n,flexblockstr * fb)
{
 os_error * err;
 char     * prev;

 err=NULL;

 prev=fb->flex__lim;

 flex__slot(n,fb);

 if(fb->flex__lim<prev+n) /* restore state and lose extra memory */
 {
  flex__slot(prev-fb->flex__lim,fb);
  err=geterror(EMEM);
 }

 return(err);
}



static void flex_demon(flexblockstr * fb)
{
 int free;

 free=fb->flex__lim-fb->flex__freep;
 free-=fb->overhead;

 if(free>0) flex__slot(-free,fb);

 remzeroevent((zerofn)flex_demon,(int)fb);

 fb->demon=0;
}


/* there is a problem because addzeroevent() may be called before */
/* events are initialised */

static os_error * flex__give(flexblockstr * fb)
{
 int free;

 if(!fb->demon)
 {
  free=fb->flex__lim-fb->flex__freep;
  if(free>fb->overhead)
  {
   if(!addzeroevent((zerofn)flex_demon,(int)fb)) fb->demon=1;
  }
 }

 return(NULL);
}


static os_error * flex__ensure(int n,flexblockstr * fb)
{
 n-=fb->flex__lim-fb->flex__freep;

 if(n<=0) return(NULL);
 else     return(flex__more(n,fb));
}



os_error * flex_alloca(flex_ptr anchor,int n,flexblockstr * fb)
{
 os_error  * err;
 flex__rec * p;

 err=NULL;


 if(n<0)
 {
  *anchor=0;
 }
 else
 {
  err=flex__ensure(sizeof(flex__rec)+roundup(n),fb);
  if(!err)
  {
   p=(flex__rec*)fb->flex__freep;
   fb->flex__freep+=sizeof(flex__rec)+roundup(n);

   p->anchor=anchor;
   p->size=n;

#ifdef FLEXEX
   p->lock=NULL;
#endif

   *anchor=p+1; /* sizeof(flex__rec), that is */
  }
 }
 return(err);
}



static os_error * flex__reanchor(flex__rec *p,int by,flexblockstr * fb)
{
 os_error * err;

 /* Move all the anchors from p upwards. This is in anticipation
    of that block of the heap being shifted. */

 err=NULL;

 while(1)
 {
  if((int)p>=(int)fb->flex__freep) break;
  if(*(p->anchor)!=p+1)
  {
   err=geterror(EMEMF);
   break;
  }

#ifdef FLEXEX
  if(p->lock) p->lock(p->handle,by);
#endif

  *(p->anchor)=((char*)(p+1))+by;
  p=(flex__rec*)(((char*)(p+1))+roundup(p->size));
 }

 return(err);
}



os_error * flex_freea(flex_ptr anchor,flexblockstr * fb)
{
 os_error  * err;

 if(*anchor)
 {
  flex__rec * p=((flex__rec*) *anchor) - 1;
  int roundsize=roundup(p->size);
  flex__rec * next=(flex__rec*)(((char*)(p+1))+roundsize);

  if(p->anchor!=anchor) err=geterror(EMEMF);
  else
  {
   err=flex__reanchor(next,-((int)sizeof(flex__rec)+roundsize),fb);

   if(!err)
   {
    memmove(p,next,fb->flex__freep-(char*)next);
    fb->flex__freep -= sizeof(flex__rec)+roundsize;
    err=flex__give(fb);
    *anchor = 0;
   }
  }
 } else err=NULL;

 return(err);
}





os_error * flex_midextenda(flex_ptr anchor,int at,int by,flexblockstr * fb)
{
 os_error  * err;
 flex__rec * p;
 flex__rec * next;

 err=NULL;

 p=((flex__rec*)*anchor)-1;

 if(p->anchor!=anchor) err=geterror(EMEMF);
 else
 if(at>p->size) err=geterror(EMEMF);
 else
 if(by<0 && (-by)>at) err=geterror(EMEMF);
 else
 if(by==0)
 {
  /* do nothing */
 }
 else
 if(by>0)
 {                                        /* extend */
  int growth=roundup(p->size+by)-roundup(p->size);
  /* Amount by which the block will actually grow. */

  err=flex__ensure(growth,fb);
  if(!err)
  {
   next=(flex__rec*)(((char*)(p+1))+roundup(p->size));

   /* The move has to happen in two parts because the moving
      of objects above is word-aligned, while the extension within
      the object may not be. */

   err=flex__reanchor(next,growth,fb);
   if(!err)
   {
    memmove(((char*)next)+roundup(growth),next,fb->flex__freep-(char*) next);

    fb->flex__freep+=growth;

    memmove(((char*)(p+1))+at+by,((char*)(p+1))+at,p->size-at);
    p->size+=by;
   }
  }
 }
 else
 { /* The block shrinks. */
  int shrinkage;

  next=(flex__rec*)(((char*)(p+1))+roundup(p->size));

  by=-by; /* a positive value now */
  shrinkage=roundup(p->size) - roundup(p->size - by);
      /* a positive value */

  memmove(((char*)(p+1))+at-by,((char*)(p+1))+at,p->size-at);
  p->size-=by;

  err=flex__reanchor(next,-shrinkage,fb);
  if(!err)
  {
   memmove(((char*)next)-shrinkage,next,fb->flex__freep-(char*) next);
   fb->flex__freep-=shrinkage;
   err=flex__give(fb);
  }
 }
 return(err);
}




os_error * flex_extenda(flex_ptr anchor,int newsize,flexblockstr * fb)
{
 flex__rec *p=((flex__rec*) *anchor)-1;
 return(flex_midextenda(anchor,p->size,newsize-p->size,fb));
}





/* called when we propose to change amount of stuff in flex block */

os_error * flex_chunka(flex_ptr anchor,int size,int chunksize,
                                                          flexblockstr * fb)
{
 os_error * err;
 flex__rec *p = ((flex__rec*) *anchor)-1;

 err=NULL;

 if((size>=p->size) || (p->size>(size+(3*chunksize)/2)))
 {
  while(chunksize)
  {
   err=flex_extenda(anchor,(size/chunksize+1)*chunksize,fb);
   if(!err) break;
   chunksize/=2;
  }
 }

 return(err);
}




os_error * flex_overheada(int overhead,flexblockstr * fb)
{
 os_error * err;
 int        delta;

 err=NULL;

 overhead=roundup(overhead);

 delta=fb->flex__lim-fb->flex__freep;
 delta=overhead-delta;

 if(delta>0)
 {
  flex__slot(delta,fb);
 }
 else
 if(delta<0)
 {
  err=flex__give(fb);
 }

 if(!err)
 {
  fb->overhead=overhead;
 }

 return(err);
}


int flex_getoverheada(flexblockstr * fb)
{
 return(fb->overhead);
}


/**************************************************************************/

os_error * flex_createarea(flexblockstr ** fbp,int bits,int max,char * name)
{
 os_error     * err;
 flexblockstr * fb;
 flexblockstr * next;
 int            handle;
 void         * base;

 fb=*fbp;

 err=Task_CreateMemoryArea(&handle,&base,max,bits,name);
 if(!err)
 {
  fb->areano=handle;
  fb->uses=0;
  fb->demon=0;
  fb->flex__freep=fb->flex__lim=fb->flex__base=(char*)base;

  if(fb!=&flexbaseblock)
  {
   next=flexbaseblock.next;
   fb->prev=&flexbaseblock;
   fb->next=next;
   flexbaseblock.next=fb;
   if(next) next->prev=fb;
  }
 }
 else
 if(fb!=&flexbaseblock)
 {
  *fbp=&flexbaseblock;
  err=NULL;
 }

 cprintf("**flex_init err=%x %s",err,err?err->errmess:"");

 return(err);
}


os_error * flex_deletearea(flexblockstr * fb)
{
 os_error     * err;
 flexblockstr * next;
 flexblockstr * prev;

 err=NULL;

 if(fb->areano && fb->uses--==0)
 {
  flex__slot(fb->flex__base-fb->flex__lim,fb);

  err=Task_DeleteMemoryArea(fb->areano);

  next=fb->next;
  prev=fb->prev;

  if(next) next->prev=prev;
  if(prev) prev->next=next;
 }

 return(err);
}



os_error * flex_finit(void)
{
 while(flexbaseblock.next) flex_deletearea(flexbaseblock.next);
 return(NULL);
}


/**************************************************************************/

/****************************************************************************/


#ifdef FLEXEX
os_error * flex_setlock(flex_ptr anchor,flex_lockfn lockfn,int handle)
{
 flex__rec *p = ((flex__rec*) *anchor)-1;
 p->lock=lockfn;
 p->handle=handle;
 return(NULL);
}
#endif



os_error * flex_size(flex_ptr anchor,int * size)
{
 os_error  * err;
 flex__rec * p=((flex__rec*)*anchor)-1;

 if(p->anchor!=anchor) err=geterror(EMEMF);
 else                  err=NULL;

 *size=p->size;

 return(err);
}


os_error * flex_swop(flex_ptr anchor1,flex_ptr anchor2)
{
 flex__rec * p1;
 flex__rec * p2;

 if(*anchor1) p1=((flex__rec*)*anchor1)-1;
 else         p1=NULL;

 if(*anchor2) p2=((flex__rec*)*anchor2)-1;
 else         p2=NULL;

 if(p1) p1->anchor=anchor2;
 if(p2) p2->anchor=anchor1;

 if(p2) *anchor1=p2+1;
 else   *anchor1=NULL;

 if(p1) *anchor2=p1+1;
 else   *anchor2=NULL;

 return(NULL);
}





os_error * flex_init(void)
{
 os_error     * err;
 flexblockstr * fbp;

 fbp=&flexbaseblock;
 
 err=flex_createarea(&fbp,0,-1,NULL);

 cprintf("post flex inti");

 return(err);
}


/**************************************************************************/


#ifdef __DEBUG

void flex_debug(flex_ptr anchor,int line)
{
 flex__rec *p = ((flex__rec*) *anchor)-1;

 dprintf(line,"anchor=%x p=%x p->anchor=%x",anchor,p,p->anchor);
}



int flex_checka(flexblockstr * fb)
{
 flex__rec * p;
 char      * q;

 p=(flex__rec *)fb->flex__base;

 while(((int)p)<((int)fb->flex__freep))
 {
  q=*(char**)(p->anchor);

  if(q<((char*)fb->flex__base) || q>((char*)fb->flex__freep))
  {
   dprintf(0,"flex pointer at %x points to illegal address.",(int)p);
   return(1);
  }

  if(*(p->anchor)!=(((char*)(p+1))))
  {
   dprintf(0,"bad flex blockat %x.",(int)p);
   return(1);
  }

 dprintf(1,"block %x len=%d.",p,p->size); 

  p=(flex__rec*)(((char*)(p + 1))+roundup(p->size));
 }

 return(0);
}


int flex_check(void)
{
 return(flex_checka(&flexbaseblock));
}








#endif


void flex_statsa(flexblockstr * fb)
{
 cprintf("base %x",(int)fb->flex__base);
 cprintf("lim  %x",(int)fb->flex__lim);
 cprintf("free %x",(int)fb->flex__freep);
}



void flex_stats(void)
{
 flex_statsa(&flexbaseblock);
}


#else // FIXEDMEM

//#define FIXDEBUG



typedef struct flex__rec
{

  int         size;      /* in bytes. Exact size of logical area. */

#ifdef FLEXEX

  flex_lockfn lock;      /* fn to call when block is moved        */
  int         handle;    /* user handle                           */
                         /* then the actual store follows.        */
#endif


#ifdef FIXDEBUG

  flex__rec * next;
  flex__rec * prev;
  int         check1;
  int         check2;
  int         check3;
#endif
} flex__rec;


#ifdef FIXDEBUG
#define TCHECK (3*sizeof(int))
#else
#define TCHECK 0
#endif


#include "malloc.h"


#ifdef NEVER

#include <crtdbg.h>

static void * xrealloc(void * ptr,size_t size)
{
 void * nptr;
 size_t oldsize;

//  cprintf("heap xr %d",HeapValidate(GetProcessHeap(),0,NULL));

// cprintf("valid %d",_CrtIsValidHeapPointer(ptr));

 nptr=malloc(size);

 if(nptr)
 {
  oldsize=_msize(ptr);
  memcpy(nptr,ptr,MIN(size,oldsize));
  free(ptr);
 }

 return(nptr);
}


#endif




static flexblockstr flexbaseblock;


#ifdef FIXDEBUG

static flex__rec * firstblock;


int flex_check(void)
{
 flex__rec * p;
 int       * q;
 int         n;

 p=firstblock;
// n=10;
 xcprintf("check start");
 while(p)
 {
//  xcprintf("size %d",p->size);
  if(p->check1!=0x12345678) xcprintf("bad check1");
  if(p->check2!=0x12345678) xcprintf("bad check2");
  if(p->check3!=0x12345678) xcprintf("bad check3");

  q=(int*)(((char*)p)+p->size-TCHECK);

  if(q[0]!=0x12345678) xcprintf("bad checkt1"); 
  if(q[1]!=0x12345678) xcprintf("bad checkt2");
  if(q[2]!=0x12345678) xcprintf("bad checkt3");

  p=p->next;
//  n--;
//  if(n<=0) break;
 }

 return(0);
}

#endif


#include <crtdbg.h>


os_error * flex_alloca(flex_ptr anchor,int n,flexblockstr * fb)
{
 char * p;
#ifdef FIXDEBUG
 int  * q;
#endif

//   cprintf("Alloc %d valid %d",n,_CrtCheckMemory());


 if(n<0) return(geterror(EMEMF));

 n+=sizeof(flex__rec)+TCHECK;
 
 p=(char*)calloc(n,1);
 if(p==NULL) return(geterror(EMEM));

 ((flex__rec*)p)->size=n;
 *anchor=(flex_ptr)(p+sizeof(flex__rec));
#ifdef FLEXEX
 ((flex__rec*)p)->lock=NULL;
#endif


#ifdef FIXDEBUG
 ((flex__rec*)p)->next=firstblock;
 ((flex__rec*)p)->prev=NULL;
 if(firstblock) firstblock->prev=((flex__rec*)p);
 firstblock=((flex__rec*)p);
 ((flex__rec*)p)->check1=0x12345678;
 ((flex__rec*)p)->check2=0x12345678;
 ((flex__rec*)p)->check3=0x12345678;
 q=(int*)(p+n-TCHECK);
 q[0]=0x12345678;
 q[1]=0x12345678;
 q[2]=0x12345678;
#endif

 return(NULL);
 USE(fb);
}




os_error * flex_freea(flex_ptr anchor,flexblockstr * fb)
{
 if(*anchor)
 {
//  cprintf("flex free");
#ifdef FIXDEBUG
  flex__rec * p;
  flex__rec * next;
  flex__rec * prev;

  p=(flex__rec*)(((char*)(*anchor))-sizeof(flex__rec));
  next=p->next;
  prev=p->prev;
  if(next) next->prev=prev;
  if(prev) prev->next=next;
  if(firstblock==p) firstblock=next;
#endif

//  cprintf("Free valid %d",_CrtCheckMemory());


  free(((char*)(*anchor))-sizeof(flex__rec));

  *anchor=NULL;
 }
 return(NULL);
 USE(fb);
}


os_error * flex_extenda(flex_ptr anchor,int newsize,flexblockstr * fb)
{
 char * mem;
 char * mem1;

// cprintf("extend");

//  cprintf("Extend %d valid %d",newsize,_CrtCheckMemory());

 if(newsize<0) return(geterror(EMEMF));

 newsize+=sizeof(flex__rec)+TCHECK;
 mem1=mem=((char*)(*anchor))-sizeof(flex__rec);

 mem=(char*)realloc(mem,newsize);
 if(mem==NULL)
 {

  return(geterror(EMEM));
 }
 else
 {
#ifdef FIXDEBUG
  flex__rec * next;
  flex__rec * prev;
  int       * q;
#endif

  ((flex__rec*)mem)->size=newsize;
  *anchor=(flex_ptr)(mem+sizeof(flex__rec));


#ifdef FLEXEX
  if(mem!=mem1)
  {
   if(((flex__rec*)mem)->lock) ((flex__rec*)mem)->lock(((flex__rec*)mem)->handle,mem-mem1);
  }
#endif

if(((flex__rec*)mem)->lock)
{
 if(mem<mem1) cprintf("!!!!XXX realloc down");
}

#ifdef FIXDEBUG
  next=((flex__rec*)mem)->next;
  prev=((flex__rec*)mem)->prev;
  if(next) next->prev=((flex__rec*)mem);
  if(prev) prev->next=((flex__rec*)mem);
  else     firstblock=((flex__rec*)mem);

  q=(int*)(mem+newsize-TCHECK);
  q[0]=0x12345678;
  q[1]=0x12345678;
  q[2]=0x12345678;
#endif
 }

 return(NULL);
 USE(fb);
}



os_error * flex_midextenda(flex_ptr anchor,int at,int by,flexblockstr * fb)
{
 char     * mem;
 char     * mem1;
 int        length;
 int        newsize;

//   cprintf("MidExtend %d %d valid %d",at,by,_CrtCheckMemory());

 mem=mem1=((char*)(*anchor))-sizeof(flex__rec);
 length=((flex__rec*)mem)->size;
 if((by+at)<0) by=-at;

 if(at<0) return(geterror(EMEMF));

// cprintf("valid %d",_CrtIsValidHeapPointer((void*)mem));
// cprintf("size %d",length);

 at+=sizeof(flex__rec);

 if(at>length) return(geterror(EMEMF));

 if(by<0)
 {
#ifdef FIXDEBUG
   flex__rec * next;
   flex__rec * prev;
   int       * q;
#endif

  newsize=length+by;

  by=-by;
  memmove(((char*)(mem))+at-by,((char*)(mem))+at,length-at-TCHECK);
  mem=(char*)realloc(mem,length-by);
  if(mem==NULL)
  {
   mem=((char*)(*anchor))-sizeof(flex__rec);
  }

  ((flex__rec*)mem)->size-=by;
  *anchor=(flex_ptr)(mem+sizeof(flex__rec));

#ifdef FLEXEX
   if(mem!=mem1)
   {
    if(((flex__rec*)mem)->lock) ((flex__rec*)mem)->lock(((flex__rec*)mem)->handle,mem-mem1);
   }
#endif

#ifdef FIXDEBUG
   next=((flex__rec*)mem)->next;
   prev=((flex__rec*)mem)->prev;
   if(next) next->prev=((flex__rec*)mem);
   if(prev) prev->next=((flex__rec*)mem);
   else     firstblock=((flex__rec*)mem);

   q=(int*)(mem+newsize-TCHECK);
   q[0]=0x12345678;
   q[1]=0x12345678;
   q[2]=0x12345678;
#endif


 }
 else
 if(by>0)
 {
  newsize=length+by;

  mem=(char*)realloc(mem,newsize);

  if(mem==NULL)
  {
   return(geterror(EMEM));
  }
  else
  {
#ifdef FIXDEBUG
   flex__rec * next;
   flex__rec * prev;
   int       * q;
#endif

   ((flex__rec*)mem)->size=newsize;
   *anchor=(flex_ptr)(mem+sizeof(flex__rec));

#ifdef FIXDEBUG
   next=((flex__rec*)mem)->next;
   prev=((flex__rec*)mem)->prev;
   if(next) next->prev=((flex__rec*)mem);
   if(prev) prev->next=((flex__rec*)mem);
   else     firstblock=((flex__rec*)mem);

   q=(int*)(mem+newsize-TCHECK);
   q[0]=0x12345678;
   q[1]=0x12345678;
   q[2]=0x12345678;
#endif

   memmove(((char*)(mem))+at+by,((char*)(mem))+at,length-at-TCHECK);

#ifdef FLEXEX
   if(mem!=mem1)
   {
    if(((flex__rec*)mem)->lock) ((flex__rec*)mem)->lock(((flex__rec*)mem)->handle,mem-mem1);
   }
#endif

   if(((flex__rec*)mem)->lock)
   {
    if(mem<mem1) cprintf("!!!!XXX realloc down");
   }
  }
 }

 return(NULL);
 USE(fb);
}


os_error * flex_chunka(flex_ptr anchor,int size,int chunksize,flexblockstr * fb)
{
 void            * mem;
 int               length;

// cprintf("Chunk %d %d valid %d",size,chunksize,_CrtCheckMemory());

 mem=*anchor;
 if(!mem)
 {
  return(flex_alloca(anchor,MAX(size,chunksize),fb));
 }
 else
 {
  flex_size(anchor,&length);

  if((size>=length) || (length>(size+chunksize)))
  {
   return(flex_extenda(anchor,(size/chunksize+1)*chunksize,fb));
  }
 }

 return(NULL);
}



os_error * flex_createarea(flexblockstr ** fbp,int bits,int max,char * name)
{
 USE(fbp);
 USE(bits);
 USE(max);
 USE(name);
 return(NULL);
}



os_error * flex_deletearea(flexblockstr * fb)
{
 USE(fb);
 return(NULL);
}



os_error * flex_finit(void)
{
 return(NULL);
}



os_error * flex_init(void)
{
/* 
   _CrtSetReportMode( _CRT_WARN, _CRTDBG_MODE_FILE );
   _CrtSetReportFile( _CRT_WARN, _CRTDBG_FILE_STDOUT );
   _CrtSetReportMode( _CRT_ERROR, _CRTDBG_MODE_FILE );
   _CrtSetReportFile( _CRT_ERROR, _CRTDBG_FILE_STDOUT );
   _CrtSetReportMode( _CRT_ASSERT, _CRTDBG_MODE_FILE );
   _CrtSetReportFile( _CRT_ASSERT, _CRTDBG_FILE_STDOUT );
*/
 
 
 return(NULL);
}



os_error * flex_swop(flex_ptr anchor1,flex_ptr anchor2)
{
 void * mem1;
 void * mem2;

 mem1=*anchor1;
 mem2=*anchor2;

 *anchor1=mem2;
 *anchor2=mem1;

 return(NULL);
}



os_error * flex_size(flex_ptr anchor,int * size)
{
 *size=((flex__rec*)(((char*)(*anchor))-sizeof(flex__rec)))->size-sizeof(flex__rec)-TCHECK;
 return(NULL);
}



os_error * flex_overheada(int overhead,flexblockstr * fb)
{
 USE(overhead);
 USE(fb);
 return(NULL);
}



int flex_getoverheada(flexblockstr * fb)
{
 return(0);
 USE(fb);
}



#ifdef FLEXEX
os_error * flex_setlock(flex_ptr anchor,flex_lockfn lockfn,int handle)
{
 flex__rec *p = ((flex__rec*) *anchor)-1;
 p->lock=lockfn;
 p->handle=handle;
 return(NULL);
}
#endif


int flex_freemem(void)
{
 return(0x7FFFFFFF);
}



void flex_statsa(flexblockstr * fb)
{
 USE(fb);
}



void flex_stats(void)
{
 flex_statsa(&flexbaseblock);
}





#endif // FIXEDMEM

/**************************************************************/



static flex_panicfn flex__panic;


os_error * flex_addpanic(flex_panicfn panic)
{
 flex__panic=panic;
 return(NULL);
}



static os_error * flex_extendb(flex_ptr anchor,int newsize,flexblockstr * fb)
{
 os_error * err=NULL;

#ifdef FLEXCHECK
 cprintf("flex_extend");
#endif


 while(1)
 {
  err=flex_extenda(anchor,newsize,fb);
  if(!err) break;

#ifdef FLEXCHECK
 cprintf("flex_extend wants to panic");
#endif

  if(flex__panic)
  {

#ifdef FLEXCHECK
 cprintf("flex_extend about to panic");
#endif

   err=flex__panic(0);
   if(err) break;
  }
  else break;
 }
 return(err);
}



static os_error * flex_allocb(flex_ptr anchor,int n,flexblockstr * fb)
{
 os_error * err=NULL;



#ifdef FLEXCHECK
 cprintf("flex_alloc");
#endif


 while(1)
 {
  err=flex_alloca(anchor,n,fb);
  if(!err) break;


#ifdef FLEXCHECK
 cprintf("flex_alloc wants to panic");
#endif


  if(flex__panic)
  {

#ifdef FLEXCHECK
 cprintf("flex_alloc about to panic");
#endif


   err=flex__panic(0);
   if(err) break;
  }
  else break;
 }
 return(err);
}










os_error * flex_alloc(flex_ptr anchor,int n)
{
 return(flex_allocb(anchor,n,&flexbaseblock));
}

os_error * flex_free(flex_ptr anchor)
{
 return(flex_freea(anchor,&flexbaseblock));
}

os_error * flex_extend(flex_ptr anchor,int newsize)
{
 return(flex_extendb(anchor,newsize,&flexbaseblock));
}

os_error * flex_midextend(flex_ptr anchor,int at,int by)
{
 return(flex_midextenda(anchor,at,by,&flexbaseblock));
}

os_error * flex_chunk(flex_ptr anchor,int size,int chunksize)
{
 return(flex_chunka(anchor,size,chunksize,&flexbaseblock));
}

os_error * flex_overhead(int overhead)
{
 return(flex_overheada(overhead,&flexbaseblock));
}

int flex_getoverhead(void)
{
 return(flex_getoverheada(&flexbaseblock));
}



void flex_checkheap(char * string) 
{
#ifdef _DEBUG
 
 int code;
 cprintf("Check memory %s valid %d",string,code=_CrtCheckMemory());

 cprintf("heap %s",HeapValidate(GetProcessHeap(),0,NULL)?"OK":"Bad");

#endif
}





