/*->c.poll */

#include <stdafx.h>

#define WM_MOUSEWHEEL                   0x020A
#define WM_MOUSEHWHEEL                  0x020E
#define WM_APPCOMMAND                   0x0319

#define FAPPCOMMAND_MASK  0xF000

#define GET_APPCOMMAND_LPARAM(lParam) ((short)(HIWORD(lParam) & ~FAPPCOMMAND_MASK))
#define GET_DEVICE_LPARAM(lParam)     ((WORD)(HIWORD(lParam) & FAPPCOMMAND_MASK))

#define GET_FLAGS_LPARAM(lParam)      (LOWORD(lParam))
#define GET_KEYSTATE_LPARAM(lParam)   GET_FLAGS_LPARAM(lParam)



#include "dde.h"

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <time.h>

#include "os.h"
#include "wimpt.h"
#include "werr.h"

#include "flex.h"

#include "alloc.h"
#include "etc.h"
#include "err.h"
#include "wos.h"
#include "fsx.h"
#include "timex.h"
#include "deb.h"
#include "osunits.h"
#include "colour.h"
#include "hourglass.h"
#include "key.h"
#include "sp.h"
#include "bits.h"
#include "xhelp.h"

#include "osmlo.h"
#include "osglue.h"
#include "oserr.h"

#include "osdrag.h"
#include "oshelp.h"
#include "oscb.h"
#include "userdraw.h"
#include "abx.h"
#include "osdde.h"
#include "task.h"



#include "poll.h"
#include "ospoll.h"


/*
  Compare

  XMapWindowPoints(hwnd,GetDesktopWindow(),&pt,1,&result);
  with
  ClientToScreen(hwnd,&pt);

  XGetWindowLong(child,GWL_ID,(long*)&m.i);
  with
  GetDlgCtrlId(child);

 (HBRUSH)(COLOR_BTNFACE+1)
  with
  GetSysColorBrush(COLOR_BTNFACE)

  GetNextWindow and GetWindow - same

  GLW_USERDATA is set with an offset of -21 (word boundary?)
  GLW_EXSTYLE  is cached

*/






/*****************************************************************************/


#define POLLENTER    -2
#define POLLEXIT     -3
#define DATAOPEN     -4
#define DATAOPENPOST -5
#define DATALOAD     -6
#define DATALOADPOST -7
#define DATALOADTYPE -8
#define DATALOADNAME -9
#define EVENT        -10
#define WINMESSAGE0  -11
#define WINMESSAGE1  -12
#define WINMASK0     -13
#define WINMASK1     -14
#define SCHEDULE     -15
#define HELPMESSAGE  -16
#define MOUSEMOVE    -17
#define ACTIVATE     -18
#define SETCURSOR    -19
#define EDITUPDATE   -20
#define EDITCARET    -21
#define SUBCLASS     -22
#define WINDOWDELETE -23
#define PARENTCLOSE  -24
#define MOUSEEVENT   -25
#define HOTKEY       -26
#define WHEEL        -27
#define APPCOMMAND   -28
#define WINMESSMASK0 -29
#define WINMESSMASK1 -30


typedef os_error * (*subfn)(uservalue userhandle,HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam,int * hit,LRESULT * code);




typedef struct eventstr
{
 int type;      /* e.g. zero poll, redraw or message */
 uservalue message;   /* sub type - message number         */
 wimp_w handle;    /* window handle */
 uservalue icon;      /* icon */

 union
 {
  zerofn     zero;
  pollfn     poll;
  redrfn     redraw;
  openfn     open;
  closefn    close;
//  ptrfn      ptr;
  clickfn    click;
  keyfn      key;
  caretfn    caret;
  activatefn activate;
  messfn     message;
  loadtypefn loadtype;
  loadnamefn loadname;
  loadfn     load;
  loadpostfn loadpost;
  eventfn    event;
//  pollmaskfn pollmask;
//  prequitfn  prequit;
  helpfn     help;
  winmessfn  winmessage;
  setcursorfn setcursor;
  editupdatefn editupdate;
  editcaretfn  editcaret;
  mouseeventfn mouseevent;
  subfn        sub;
  hotkeyfn     hot;
  wheelfn      wheel;
		appcommandfn appcommand;

 } fn;

} eventstr;


#define EVMIN   64
#define EVCHUNK (EVMIN*sizeof(eventstr))


static eventstr * eventstack;
static int        eventn;


typedef struct eventscanstr
{
 int                   eventscan;
 struct eventscanstr * next;
 struct eventscanstr * prev;
} eventscanstr;


static eventscanstr * eventscanlist;


static int threadlevel;

static int pollflags;


/*****************************************************************************/


void poll_check(char * where)
{
#ifdef _DEBUG

 cprintf("poll check %s %d",where,eventn);

#endif
}


static int findblock(int type,uservalue message,wimp_w handle,uservalue icon,zerofn fn)
{
 int        i;

 if(eventstack)
 {
  for(i=0;i<eventn;i++)
  {
   if((type==-1    || eventstack[i].type==type) &&
      (message==-1 || eventstack[i].message==message) &&
      (handle==((wimp_w)-1)  || eventstack[i].handle==handle) &&
      (icon==((uservalue)-1)    || eventstack[i].icon==icon) &&
      (!fn         || eventstack[i].fn.zero==fn)) return(i);
  }
 }
 return(-1);
}


static os_error * addblock(int type,uservalue message,wimp_w handle,uservalue icon,zerofn fn,
                                                             int * hit)
{
 os_error * err;
 int        i;

// cprintf("add block=%d",type);

 if(hit) *hit=0;

 if(!eventstack)
 {
  return(geterror(ENOEVENT));
 }

// cprintf("search message");

 for(i=0;i<eventn;i++)
 {
  if(eventstack[i].type==type &&
     eventstack[i].message==message &&
     eventstack[i].handle==handle &&
     eventstack[i].icon==icon &&
     eventstack[i].fn.zero==fn)
  {
   return(NULL);
  }
 }


//cprintf("flexchunk");

 err=flex_chunk((flex_ptr)&eventstack,(eventn+1)*sizeof(eventstr),EVCHUNK);
 if(!err)
 {
//  cprintf("adding message");
  eventstack[eventn].type=type;
  eventstack[eventn].message=message;
  eventstack[eventn].handle=handle;
  eventstack[eventn].icon=icon;
  eventstack[eventn].fn.zero=fn;
  eventn++;
  if(hit) *hit=1;
 }

// cprintf("eventn+=%d",eventn);

 return(err);
}


static os_error * remblock(int type,uservalue message,wimp_w handle,uservalue icon,zerofn fn,
                                                        int * hit)
{
 int            i;
 int            j;
 eventscanstr * es;

 if(hit) *hit=0;

 if(!eventstack)
 {
  return(NULL);
 }

#ifdef _DEBUG

 int eventn0=eventn;

#endif


 for(i=0;i<eventn;i++)
 {
  if(eventstack[i].type==type &&
     eventstack[i].message==message &&
     eventstack[i].handle==handle &&
     eventstack[i].icon==icon &&
     eventstack[i].fn.zero==fn)
  {
   for(j=i;j<(eventn-1);j++) eventstack[j]=eventstack[j+1];
   eventn--;

   es=eventscanlist;
   while(es)
   {
    if(es->eventscan>=i) es->eventscan--;
    es=es->next;
   }

   if(hit) *hit=1;
   break;
  }
 }

#ifdef _DEBUG
 if(i>=eventn0) cprintf("remblock missed type %d message %d",type,message);

#endif


// cprintf("eventn-=%d",eventn);  
// cprintf("last=%d",eventstack[eventn-1].type);

 return(NULL);
}


static void starteventscan(eventscanstr * es)
{
 es->eventscan=-1;

 if(eventscanlist) eventscanlist->prev=es;
 es->next=eventscanlist;
 es->prev=NULL;
 eventscanlist=es;
}


static void restarteventscan(eventscanstr * es)
{
 es->eventscan=-1;
}



static void endeventscan(eventscanstr * es)
{
 eventscanstr * prev;
 eventscanstr * next;

 if(eventscanlist==es) eventscanlist=es->next;
 else
 {
  prev=es->prev;
  next=es->next;

  if(prev) prev->next=next;
  if(next) next->prev=prev;
 }
}




static int nextevent(eventscanstr * es,int type,uservalue message,
                                                wimp_w handle,uservalue icon)
{
 int i;

 es->eventscan++;


 for(i=es->eventscan;i<eventn;i++)
 {
  if((type==-1    || eventstack[i].type==type) &&
     (message==-1 || eventstack[i].message==message) &&
     (handle==((wimp_w)-1)  || eventstack[i].handle==handle) &&
     (icon==-1    || eventstack[i].icon==icon))
  {
//   cprintf("event hit type %d",eventstack[i].type);
   es->eventscan=i;
   return(i);
  }
 }

 return(-1);
}


/*****************************************************************************/

static int dopollidle;
static int doschedule;
       int zerotime;
       int keytime;
static UINT_PTR scheduletimer=0;

static os_error * checkschedule(int action);



static VOID CALLBACK scheduleproc(HWND hwnd,UINT uMsg,UINT_PTR idEvent,DWORD dwTime)
{
 threadlevel++;
 checkschedule(1);
 threadlevel--;
 USE(hwnd);
 USE(uMsg);
 USE(idEvent);
 USE(dwTime);
}
 

// all too likely this will recurse, e.g. timer adds another timer.

static os_error * checkschedule(int action)
{
 int          mono;
 int          event;
 wimp_w       ehandle;
 zerofn       efn;
 int          hit;
 eventscanstr es;
 int          timeout;
 int          delay;
// static int   recurse=0;

// if(recurse) return(NULL);
// recurse++;

 mono=zerotime=monotonic();  
 timeout=0;

 if(doschedule)
 {
  starteventscan(&es);

  while((event=nextevent(&es,SCHEDULE,-1,(wimp_w)-1,-1))>=0)
  {
   if(tickdiff(eventstack[event].icon,mono)<=0 && action)
   {
    ehandle=eventstack[event].handle;
    efn=eventstack[event].fn.zero;

    remblock(SCHEDULE,0,ehandle,eventstack[event].icon,efn,&hit);
    doschedule--;    /* here to stop recursion */

    efn((uservalue)ehandle);

    event=es.eventscan;
   }
   else
   {
    if(timeout==0 || tickdiff(eventstack[event].icon,timeout)<0) timeout=eventstack[event].icon;
   }
  }
  endeventscan(&es);
 }

  
 if(scheduletimer)
 {
  KillTimer(NULL,scheduletimer);
  scheduletimer=0;
 }

 if(doschedule)
 {
  if(timeout<=mono) delay=1;
  else              delay=(timeout-mono)*10;
  scheduletimer=SetTimer(NULL,0,delay,scheduleproc);
 }


// recurse--;

 return(NULL);
}


// the problem with this is calling checkschedule() is likely to cause
// recursion for example one redraw inside another
// e.g. font cache adds an event during redraw and lets caret update go off
// use action parameter to prevent this.

os_error * addscheduledevent(zerofn fn,uservalue userhandle,int delay)
{
 os_error * err;
 int        hit;
 int        event;
 int        time;

 err=NULL;

 time=monotonic()+delay;

 event=findblock(SCHEDULE,0,(wimp_w)userhandle,-1,fn);

 if(event>=0) eventstack[event].icon=time;
 else
 {
  err=addblock(SCHEDULE,0,(wimp_w)userhandle,time,fn,&hit);
  if(!err && hit) doschedule++;
 }

 checkschedule(0);

 return(err);
}



/* unique in that we don't know the 'icon' value   */
/* so we do a search for first match               */
/* things would be easier if remblock matched -1's */


os_error * remscheduledevent(zerofn fn,uservalue userhandle)
{
 os_error * err;
 int        hit;
 int        event;

 err=NULL;

 event=findblock(SCHEDULE,0,(wimp_w)userhandle,-1,fn);
 if(event>=0) 
 {
  err=remblock(SCHEDULE,0,(wimp_w)userhandle,eventstack[event].icon,fn,&hit);
  if(!err && hit)
  {
   doschedule--;
   checkschedule(0);
  }
 }

 return(err);
}


os_error * addzeroevent(zerofn fn,uservalue userhandle)
{
 os_error * err;
 int        hit;

 err=addblock(wimp_ENULL,0,(wimp_w)userhandle,0,fn,&hit);
 if(!err && hit) dopollidle++;

 return(err);
}

os_error * remzeroevent(zerofn fn,uservalue userhandle)
{
 os_error * err;
 int        hit;

 err=remblock(wimp_ENULL,0,(wimp_w)userhandle,0,fn,&hit);
 if(!err && hit) dopollidle--;

 return(err);
}


static void zero(void)
{
 int          mono=monotonic();
 int          event;
 eventscanstr es;

 zerotime=mono;

 if(dopollidle)
 {
  starteventscan(&es);

  while((event=nextevent(&es,wimp_ENULL,-1,(wimp_w)-1,-1))>=0)
  {
   eventstack[event].fn.zero((uservalue)eventstack[event].handle);
  }

  endeventscan(&es);
 }

}

/*****************************************************************************/

/* static int polltime; */



int polllevel(void)
{
 return(threadlevel);
}


static void pollexit(void)
{
 int event;
 eventscanstr es;

 starteventscan(&es);

 while((event=nextevent(&es,POLLEXIT,-1,(wimp_w)-1,-1))>=0)
 {
  eventstack[event].fn.poll((uservalue)eventstack[event].handle);
 }

 endeventscan(&es);

/* dprintf(2,"polltime=%d",clock()-polltime); */

 threadlevel--;
}


static void pollenter(void)
{
 int event;
 eventscanstr es;

/* polltime=clock(); */

 threadlevel++;

 starteventscan(&es);

 while((event=nextevent(&es,POLLENTER,-1,(wimp_w)-1,-1))>=0)
 {
  eventstack[event].fn.poll((uservalue)eventstack[event].handle);
 }

 endeventscan(&es);
}

os_error * addpollexit(pollfn fn,uservalue userhandle)
{
 os_error * err;
 err=addblock(POLLEXIT,0,(wimp_w)userhandle,0,fn,NULL);
 return(err);
}

os_error * rempollexit(pollfn fn,uservalue userhandle)
{
 os_error * err;
 err=remblock(POLLEXIT,0,(wimp_w)userhandle,0,fn,NULL);
 return(err);
}



os_error * addpollenter(pollfn fn,uservalue userhandle)
{
 os_error * err;
 err=addblock(POLLENTER,0,(wimp_w)userhandle,0,fn,NULL);
 return(err);
}

os_error * rempollenter(pollfn fn,uservalue userhandle)
{
 os_error * err;
 err=remblock(POLLENTER,0,(wimp_w)userhandle,0,fn,NULL);
 return(err);
}









#ifdef NEVER

// pollmask, attempt to let apps look at messages first


static int pollmask(MSG * msg)
{
 int event;
 int claimed;
 eventscanstr es;

 claimed=0;
 starteventscan(&es);

 while((event=nextevent(&es,POLLMASK,-1,-1,-1))>=0)
 {
  claimed|=eventstack[event].fn.pollmask(eventstack[event].handle,msg);
 }

 endeventscan(&es);

 return(claimed);
}


os_error * addpollmask(pollmaskfn fn,uservalue userhandle)
{
 os_error * err;
 err=addblock(POLLMASK,0,userhandle,0,(zerofn)fn,NULL);
 return(err);
}


os_error * rempollmask(pollmaskfn fn,uservalue userhandle)
{
 os_error * err;
 err=remblock(POLLMASK,0,userhandle,0,(zerofn)fn,NULL);
 return(err);
}


#endif



/*****************************************************************************/

os_error * event(int eventn,void * data,int data1)
{
 os_error   * err;
 int          event;
 eventscanstr es;

 err=NULL;

 starteventscan(&es);

// cprintf("eventn=%x",eventn);

 while((event=nextevent(&es,EVENT,eventn,(wimp_w)-1,-1))>=0)
 {
/*  dprintf(0,"eventx"); */
  err=eventstack[event].fn.event(eventn,eventstack[event].handle,data,data1);
  if(err) break;
 }

 endeventscan(&es);

 return(err);
}

os_error * addevent(int eventn,eventfn fn,uservalue userhandle)
{
 os_error * err;
 err=addblock(EVENT,eventn,(wimp_w)userhandle,0,(zerofn)fn,NULL);
 return(err);
}

os_error * remevent(int eventn,eventfn fn,uservalue userhandle)
{
 os_error * err;
 err=remblock(EVENT,eventn,(wimp_w)userhandle,0,(zerofn)fn,NULL);
 return(err);
}

/*****************************************************************************/

static int redraw(MSG * msg)
{
 int          event;
 wimp_w       handle;
 eventscanstr es;
 int          hit;

 handle=(wimp_w)msg->hwnd;

 starteventscan(&es);
 hit=0;

 while((event=nextevent(&es,wimp_EREDRAW,-1,handle,-1))>=0)
 {
  report(eventstack[event].fn.redraw(handle,eventstack[event].icon));
  hit=1;
 }

 endeventscan(&es);

 return(!hit);
}


os_error * addredrawevent(redrfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(wimp_EREDRAW,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


os_error * remredrawevent(redrfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(wimp_EREDRAW,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}



/****************************************************************************/

static int wclose(MSG * msg)
{
 int         event;
 wimp_w      handle;
 int         hit;
 eventscanstr es;

 handle=(wimp_w)msg->hwnd;
 starteventscan(&es);
 hit=0;

// cprintf("wclose.... %x",handle);

 if((event=nextevent(&es,wimp_ECLOSE,-1,handle,-1))>=0)
 {
//  cprintf("hit");
  hit=1;
  report(eventstack[event].fn.close(handle,eventstack[event].icon));
 }

 endeventscan(&es);

 if(hit==0)  // problem is that Windows will destroy this window now unless we do something
 {           // CloseWindow() minimizes the window
  pollclosewindow(handle);
  ShowWindow((HWND)handle,SW_HIDE);
  hit=1;
 }

 return(!hit);
}






os_error * addcloseevent(closefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(wimp_ECLOSE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remcloseevent(closefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;

// cprintf("remclosevent %x %x %x",fn,handle,userhandle);
 err=remblock(wimp_ECLOSE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


/**************************************************************************/



static int wclosechild(wimp_w handle)
{
 int         event;
 int         hit;
 eventscanstr es;

 starteventscan(&es);
 hit=0;

// cprintf("wclose.... %x",handle);

 if((event=nextevent(&es,PARENTCLOSE,-1,handle,-1))>=0)
 {
//  cprintf("hit");
  hit=1;
  report(eventstack[event].fn.close(handle,eventstack[event].icon));
 }

 endeventscan(&es);

 return(!hit);
}




os_error * addparentcloseevent(closefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(PARENTCLOSE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


os_error * remparentcloseevent(closefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(PARENTCLOSE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}



/***************************************************************************/




static int wdeletechild(wimp_w handle)
{
 int         event;
 int         hit;
 eventscanstr es;

 starteventscan(&es);
 hit=0;

// cprintf("wclose.... %x",handle);

 if((event=nextevent(&es,WINDOWDELETE,-1,handle,-1))>=0)
 {
//  cprintf("hit");
  hit=1;
  report(eventstack[event].fn.close(handle,eventstack[event].icon));
 }

 endeventscan(&es);

 return(!hit);
}


os_error * adddeleteevent(closefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(WINDOWDELETE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


os_error * remdeleteevent(closefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(WINDOWDELETE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}



/***************************************************************************/


#ifdef NEVER

/*

static int pointerevent(MSG * msg)
{
 int          event;
 int          handle;
 eventscanstr es;
 int          enter=0;

 handle=(int)msg->hwnd;
 starteventscan(&es);

 while((event=nextevent(&es,wimp_EPTRLEAVE,-1,handle,-1))>=0)
 {
  report(eventstack[event].fn.ptr(handle,eventstack[event].icon,enter));
 }

 endeventscan(&es);
}

*/


os_error * addpointerevent(ptrfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(wimp_EPTRLEAVE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * rempointerevent(ptrfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(wimp_EPTRLEAVE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


#endif


/********************************************************************************/



static void mousemove(wimp_mousestr * m)
{
 int          event;
 wimp_w       handle;
 eventscanstr es;

 handle=m->w;
 starteventscan(&es);

//	cprintf("mouse move %x",handle);

 while((event=nextevent(&es,MOUSEMOVE,-1,handle,-1))>=0)
 {
  report(eventstack[event].fn.click(handle,eventstack[event].icon,m));
 }

 endeventscan(&es);
}




os_error * addmousemoveevent(clickfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(MOUSEMOVE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remmousemoveevent(clickfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(MOUSEMOVE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


/********************************************************************* **********/


// signal any mouse click event - e.g. for removing menus


static int   mouseeventcount;
static HHOOK mouseeventhook;



static void mouseevent(HWND hwnd,WPARAM wParam,LPARAM lParam,int * ack)
{
 int          event;
 eventscanstr es;

//	cprintf("mouseevent");

 *ack=0;                // let the event be passed on

 starteventscan(&es);

 while((event=nextevent(&es,MOUSEEVENT,-1,(wimp_w)-1,-1))>=0)
 {
//			cprintf("event handle %x",eventstack[event].handle);
  report(eventstack[event].fn.mouseevent(eventstack[event].handle,eventstack[event].icon,(wimp_w)hwnd,lParam,wParam,ack));
 }

 endeventscan(&es);
}



LRESULT CALLBACK MouseProc(int nCode,      // hook code
                           WPARAM wParam,  // message identifier
                           LPARAM lParam   // mouse coordinates
)
{
 int ack;

 mouseevent(((MOUSEHOOKSTRUCT*)lParam)->hwnd,wParam,lParam,&ack);

 if(nCode>=0)
 {
  CallNextHookEx(mouseeventhook,      // handle to current hook
                        nCode,      // hook code passed to hook procedure
                        wParam,  // value passed to hook procedure
                        lParam   // value passed to hook procedure
  );

  return(ack);
 }
 else
 {
 	return(CallNextHookEx(mouseeventhook,      // handle to current hook
                        nCode,      // hook code passed to hook procedure
                        wParam,  // value passed to hook procedure
                        lParam   // value passed to hook procedure
        ));
 }

}




static os_error * enablemouseevent(int enable)
{
	if(enable)
	{
  mouseeventhook=SetWindowsHookEx(
  WH_MOUSE,        // hook type
  MouseProc,     // hook procedure
  NULL,    // handle to application instance
  GetCurrentThreadId()   // thread identifier
  );
	}
	else
	{
		UnhookWindowsHookEx(mouseeventhook);   // handle to hook procedure
	}

	return(NULL);
}





os_error * addmouseevent(mouseeventfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(MOUSEEVENT,0,handle,userhandle,(zerofn)fn,NULL);
	if(!err)
	{
		if(mouseeventcount==0) err=enablemouseevent(1);
		if(!err) mouseeventcount++;
	}
 return(err);
}


os_error * remmouseevent(mouseeventfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(MOUSEEVENT,0,handle,userhandle,(zerofn)fn,NULL);
	if(!err)
	{
		if(mouseeventcount==1) err=enablemouseevent(0);
		if(!err) mouseeventcount--;
	}
 return(err);
}




/********************************************************************************/

// return 1 if we do something

static int wmsetcursor(MSG * msg)
{
 int           event;
 eventscanstr  es;
 wimp_mousestr m;
 int           setm;
 wimp_w        handle;
 int           code;
 int           hitcode;

 handle=(wimp_w)msg->hwnd;
 hitcode=LOWORD(msg->lParam);
 setm=0;
 code=0;

 if(hitcode==HTCLIENT)
 {
  starteventscan(&es);

  while((event=nextevent(&es,SETCURSOR,-1,handle,-1))>=0)
  {
   if(!setm)
   {
    wimp_get_point_info(&m);
    setm=1;
   }
  
   report(eventstack[event].fn.setcursor(handle,eventstack[event].icon,&m,&code));
  }

  endeventscan(&es);
 }

 return(code);
}


os_error * addsetcursorevent(setcursorfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(SETCURSOR,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remsetcursorevent(setcursorfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(SETCURSOR,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}



/********************************************************************************/

static void decodeclick(wimp_mousestr * m)
{
 int          event;
 wimp_w       handle;
 eventscanstr es;

// cprintf("decode click %d eventn=%d",m->w,eventn);

 handle=m->w;
 starteventscan(&es);

/* cprintf("type=%d mesaage=%d handle=%d icon=%d",eventstack[0].type,
     eventstack[0].message,
     eventstack[0].handle,
     eventstack[0].icon);  */



 while((event=nextevent(&es,wimp_EBUT,-1,handle,-1))>=0)
 {
  report(eventstack[event].fn.click(handle,eventstack[event].icon,m));
  break;
 }

 endeventscan(&es);
}

os_error * addclickevent(clickfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;

// cprintf("addclick %d",handle);
 err=addblock(wimp_EBUT,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remclickevent(clickfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(wimp_EBUT,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

/***************************************************************************/

#ifdef NEVER

static int  mfnncstate;
static int  mfnnclastbuttons;


   case WM_NCLBUTTONDBLCLK:
			                        mfnnclastbuttons=0x400;
			                        mfnncstate=0;
                           return(1);
   		                      break;

     case WM_NCLBUTTONDOWN:
				                       mfnnclastbuttons=0x4;
				                       mfnncstate=MSDOWN;
                           return(1);
  		                       break;

       case WM_NCLBUTTONUP:
  			                      mfnncstate=0;
                           return(1);
					                      break;
 
       case WM_NCMOUSEMOVE:
					                      if(mfnncstate==MSDOWN) mfnncstate=0;
                           return(1);
					                      break;

   case WM_NCRBUTTONDBLCLK:
			                      		mfnnclastbuttons=0x100;
						                     mfnncstate=0;
                           return(1);
						                     break;

     case WM_NCRBUTTONDOWN:
						                     mfnnclastbuttons=0x1;
						                     mfnncstate=MSDOWN;
                           return(1);
						                     break;

       case WM_NCRBUTTONUP:
					                      mfnncstate=0;
                           return(1);
					                      break;
 


    case WM_NCRBUTTONDBLCLK:
                            code=mousefn(hwnd,message,wParam,lParam);
                            //XDefWindowProc(hwnd,WM_NCLBUTTONDBLCLK,wParam,lParam,(LONG*)&code);
                            //code=0;
                            break;

    
      case WM_NCRBUTTONDOWN:
                            code=mousefn(hwnd,message,wParam,lParam);
                            // Use GetCapture()...
                            //XDefWindowProc(hwnd,WM_NCLBUTTONDOWN,wParam,lParam,(LONG*)&code);
                            //code=0;
                            break;

      
        case WM_NCRBUTTONUP:
                            code=mousefn(hwnd,message,wParam,lParam);
                            //XDefWindowProc(hwnd,WM_NCLBUTTONUP,wParam,lParam,(LONG*)&code);
                            //code=0;
                            break;

          case WM_NCHITTEST:
          case WM_SETCURSOR:
                            break;


    case WM_NCLBUTTONDBLCLK:
      case WM_NCLBUTTONDOWN:
        case WM_NCLBUTTONUP:
        case WM_NCMOUSEMOVE:
                            break;


#endif

/******************************************************************************/

static dragfn  dfn;              // at end of drag
static clickfn dufn;             // on mouse moves
static wimp_w  draghandle;
static uservalue draguserhandle;
static int     dragtype;


#define MSDOWN 0x1
#define MSDRAG 0x2


static int  mfnstate;
static wimp_mousestr savemouse;

static HWND mfndraghandle;





static void decodedrag(wimp_box * box)
{

// cprintf("decodedrag");

 if(dfn) report(dfn(draghandle,draguserhandle,box));
 dfn=NULL;    /* to stop colour picker drags causing trouble */
 dufn=NULL;
 draghandle=0;
}



os_error * wimp_drag_box(wimp_dragstr * d)
{
 os_error * err;
 HWND       hprev;

 err=NULL;

 if(d==(wimp_dragstr*)-1)
 {
//  draghandle=0;
//  mfnstate=0;       done in wmcap
  if(mfnstate==MSDRAG)  // can't just always ReleaseCap because it may be used for other things
  {
   mfnstate=0;          // prevent wmcap doing anything
   err=XReleaseCapture();
   drageffectend(NULL);
  }
 }
 else
 {
  mfnstate=0;          // prevent wmcap doing anything
  
  mfndraghandle=(HWND)d->window;
  dragtype=d->type;


  if(d->type==wimp_MOVE_WIND)
  {
   XReleaseCapture();  // because, drags launched from abx controls don't work otherwise
   SendMessage((HWND)d->window,WM_SYSCOMMAND,SC_MOVE+2,0);
  }
  else
  {
   err=XSetCapture(mfndraghandle,&hprev);
   if(!err)
   {
    drageffectstart(d);
    mfnstate=MSDRAG;
   }
//   cprintf("drag started %x %s",mfndraghandle,err?err->errmess:"");
  }
 }

 return(err);
}



static void mfndragclear(void)
{
 wimp_box box;
// cprintf("dragclear");

 if(mfnstate==MSDRAG)
 {
  XReleaseCapture();
  drageffectend(&box);
  decodedrag(&box);  // only need this here or in wmcapture
 }
 else
 if(mfnstate==MSDOWN)
 {
  XReleaseCapture();
//  cprintf("down clear");
 }
}


static void mfndragprocess(int x,int y)
{
 drageffectprocess(x,y);
}



// return 0 if process message

static int wmcapture(MSG * msg)
{
 wimp_box box;

// cprintf("wmcapture %d",mfnstate);

 if(msg->hwnd==mfndraghandle && mfnstate==MSDRAG)
 {
// decodedrag(); //  only need this here or in mfndragclear
// except in some situations there may be a wmcap message
// that does not originate in the program
// also problem of us setting off another capture e.g. in autoscroll

  drageffectend(&box);
  decodedrag(&box);
  
  mfnstate=0;
  return(0);
 }
 else
 if(mfnstate==MSDOWN)
 {
  mfnstate=0;
  return(0);
 }

 return(1);
}



static int mousecommand(HWND hwnd,HWND hwndCtl,int wID)
{
 os_error      * err;
 wimp_mousestr   m;
 POINT           pt;

 err=XGetCursorPos(&pt);
 if(!err)
 {
  m.x=xcoordpix2os(pt.x);
  m.y=ycoordpix2os(pt.y);

  m.w=(wimp_w)hwnd;
  m.i=wID;
  m.bbits=(wimp_bbits)0x4;

//  cprintf("mouse command %d",wID);

  decodeclick(&m);
 }

 return(0);

 USE(hwndCtl);
}


// the key thing is that while a state is set we have to capture the mouse
// could do it without an MDOWN state by looking at mouse buttons on mouse move
// it matters that a drag starts from the coordinates of the preceding click





static int mousefn(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam)
{
 int           xpos;
 int           ypos;
 wimp_mousestr m;
 POINT         pt;
 POINT         pt2;
 HWND          child;
 HWND          hprev;
 HWND          mwnd;


 pt.x=xpos=(short)LOWORD(lParam);
 pt.y=ypos=(short)HIWORD(lParam);
 pt2=pt;

 XClientToScreen(hwnd,&pt2);

 m.x=xcoordpix2os(pt2.x);
 m.y=ycoordpix2os(pt2.y);

 if(mfnstate==MSDRAG)          // mouse may not be over hwnd
 {
  mwnd=WindowFromPoint(pt2);
  hprev=mwnd;
  m.w=(wimp_w)mwnd;
 }
 else
 {
  m.w=(wimp_w)hwnd;
  hprev=hwnd;
 }

 while(1)
 {
  child=ChildWindowFromPoint(hprev,pt);
  if(child && child!=hprev)
  {
   hprev=child;
   XGetDlgCtrlID(hprev,&m.i);
  }
  else
  {
   m.i=-1;
   break;
  }
 }

// cprintf("mousefn i=%d message %x hwnd=%x",m.i,message,hwnd);


 switch(message)
 {
     case WM_LBUTTONDBLCLK:
                           wimp_checkfocus(hwnd);
					                      m.bbits=wimp_DOUBLELEFT;
					                      mfnstate=0;
                           savemouse=m;
					                      decodeclick(&m);
					                      break;

       case WM_LBUTTONDOWN://cprintf("ldown %x",hwnd);
                           wimp_checkfocus(hwnd);
                           if(mfnstate==MSDRAG)
                           {
                            // possible a drag was started by right down we have to clear drag
                            // also drag ended by click (as opposed to release)
                            mfndragclear(); 
				                        mfnstate=0; 
                           }
                           else
                           {
                            mfndragclear(); 
					                       m.bbits=wimp_BLEFT;
                            savemouse=m;
					                       mfnstate=MSDOWN;
                            XSetCapture(hwnd,&hprev);
					                       decodeclick(&m);
                           }
					                      break;

         case WM_LBUTTONUP://cprintf("lup");
			                       	mfndragclear();
				                       mfnstate=0;
					                      break;
 
         case WM_MOUSEMOVE:
//                           cprintf("move state=%d",mfnstate);
					                      if(mfnstate==MSDOWN)
                           {
                            mfndragclear();
					                       mfnstate=0;
                            m=savemouse;
					                       if(m.bbits==wimp_BLEFT)  m.bbits=wimp_BDRAGLEFT;
					                       else
					                       if(m.bbits==wimp_BRIGHT) m.bbits=wimp_BDRAGRIGHT;
//                            cprintf("gen drag");
					                       decodeclick(&m);
                           }
					                      else
                           if(mfnstate==MSDRAG)
                           {
                            mfndragprocess(pt.x,pt.y);
                            if(dufn) dufn(draghandle,draguserhandle,&m);
                           }
                           else
                           {
                            mousemove(&m);
                           }
					                      break;


     case WM_RBUTTONDBLCLK:
                           wimp_checkfocus(hwnd);
                           m.bbits=wimp_DOUBLERIGHT;
                           savemouse=m;
						                     mfnstate=0;
           	 			           decodeclick(&m);
						                     break;

       case WM_RBUTTONDOWN:
                           wimp_checkfocus(hwnd);
                           if(mfnstate==MSDRAG)
                           {
                            // possible a drag was started by right down we have to clear drag
                            // also drag ended by click (as opposed to release)
                            mfndragclear();
                            mfnstate=0;
                           }
                           else
                           {
                            mfndragclear();
			                         m.bbits=wimp_BRIGHT;
                            savemouse=m;
			                         mfnstate=MSDOWN;
                            XSetCapture(hwnd,&hprev);
	 		                        decodeclick(&m);
                           }
			                        break;

         case WM_RBUTTONUP:// cprintf("rbutton up");
					                      mfndragclear();
					                      mfnstate=0;
					                      break;
 }

 return(0);
 USE(wParam);
}



static os_error * mouseclose(wimp_w handle)
{
// cprintf("mouse close state %d handles %d",mfnstate,savemouse.w==handle);

 if(mfnstate && savemouse.w==handle)
 {
 	mfndragclear();
  mfnstate=0;
 }

 return(NULL);
}


os_error * startdrag(wimp_w handle,uservalue userhandle,dragfn fn)
{
 dfn=fn;
 dufn=NULL;
 draghandle=handle;
 draguserhandle=userhandle;
 return(NULL);
}


os_error * startdrag2(wimp_w handle,uservalue userhandle,dragfn fn,clickfn fn2)
{
 dfn=fn;
 dufn=fn2;
 draghandle=handle;
 draguserhandle=userhandle;
 return(NULL);
}


wimp_w getdraghandle(void)
{
 return(mfnstate==MSDRAG?draghandle:0);
}


/***************************************************************************/


// WM_CHAR and WM_KEYDOWN
// also can be called by WM_COMMAND

static int keypressskip;

static int keypress(MSG * msg,int subclass)
{
 int           event;
 wimp_w        handle;
 int           icon;
 int           key[4];
 eventscanstr  es;
 wimp_caretstr c;

 if(keypressskip) return(1);

 key[KEYTIME]=keytime=monotonic();
 key[KEYDATA]=msg->lParam;              
 key[KEYFLAGS]=subclass?KEYSUB:0;

	if(msg->message==WM_UNICHAR)
	{
		cprintf("wm_unichar nochar %d",msg->wParam==UNICODE_NOCHAR);

		if(msg->wParam==UNICODE_NOCHAR)
		{
			return(TRUE);
		}
	}


 if(msg->message==WM_CHAR || msg->message==WM_UNICHAR || msg->message==WM_IME_CHAR)
 {
  cprintf("wmchar %x",msg->wParam);
//  key[KEYCODE]=(TCHAR)msg->wParam; // problem with this if using UNICODE window in ANSI build
  key[KEYCODE]=msg->wParam;
 }
 else
 if(msg->message==WM_KEYDOWN || msg->message==WM_SYSKEYDOWN)
 {
//  cprintf("keydown %x",msg->wParam);
  key[KEYCODE]=msg->wParam;
  key[KEYCODE]|=KEYVIRT;
  if(GetKeyState(VK_SHIFT) & 0x8000)    key[KEYCODE]|=KEYVSHIFT;
  if(GetKeyState(VK_CONTROL)  & 0x8000) key[KEYCODE]|=KEYVCTRL;
  if(GetKeyState(VK_MENU)  & 0x8000)    key[KEYCODE]|=KEYVALT;
  if(GetKeyState(VK_LWIN)  & 0x8000)    key[KEYCODE]|=KEYVWIN;
  if(GetKeyState(VK_RWIN)  & 0x8000)    key[KEYCODE]|=KEYVWIN;

//  cprintf("keystate %x %x %x",GetAsyncKeyState(VK_SHIFT),GetAsyncKeyState(VK_CONTROL),GetAsyncKeyState(VK_MENU));
 }
 
 handle=(wimp_w)msg->hwnd;

 wimp_get_caret_pos(&c);
 icon=c.i;


// cprintf("wm key %x handle %x",key[0],handle);

 {
  starteventscan(&es);

  while((event=nextevent(&es,wimp_EKEY,-1,handle,-1))>=0)
  {
   report(eventstack[event].fn.key(handle,eventstack[event].icon,icon,key));
   if(key[KEYCODE]==-1) break;
  }

  endeventscan(&es);
 }

// cprintf("ret key %x",key[KEYCODE]);

 return(key[KEYCODE]!=-1);				// return true if we do nothing
}



os_error * addkeyevent(keyfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(wimp_EKEY,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remkeyevent(keyfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(wimp_EKEY,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}



/*****************************************************************************/

// called on WM_SHOWWINDOW, WM_MOVE, WM_SIZE


static int wopen(MSG * msg)
{
 int          event;
 wimp_w       handle;
 int          hit;
 eventscanstr es;
 int          code;
 wimp_wstate  winds;
 wimp_openstr2 open2;
 int          skip;

 handle=(wimp_w)msg->hwnd;

 XDefWindowProc(msg->hwnd,msg->message,msg->wParam,msg->lParam,(LRESULT*)&code);

 scrollcheck(handle,&skip);								// has to be executed
 if(skip) return(1);

 // problem is the window actually open, does the window actually exist yet?
 // only want open events from external opening of the window
 // skip open event if window is being opened by our code

 wimp_get_wind_state(handle,&winds);
 if(winds.flags & wimp_EXISTS)
 {
  if(!(winds.flags & wimp_OPENING))
  {
   starteventscan(&es);
   hit=0;
   open2.o=winds.o;
   open2.reason=msg->message;
   open2.lparam=msg->lParam;
   open2.wparam=msg->wParam;

   while((event=nextevent(&es,wimp_EOPEN,-1,handle,-1))>=0)
   {
    hit=1;
    report(eventstack[event].fn.open(handle,eventstack[event].icon,(wimp_openstr*)&open2));
   }

   endeventscan(&es);
  }
 }

 return(1);
}


// call on WM_SIZING

/*
static int sizewindow(MSG * msg)
{
 return(sizecheck(msg->hwnd,msg->wParam,(LPRECT)msg->lParam));
}

*/

static int wmgetminmax(MSG * msg)
{

 return(getminmax(msg->hwnd,(LPMINMAXINFO)msg->lParam));


}




// return 1 if we do something

#ifdef NEVER

static int wmmoving(MSG * msg)
{
 return(wimp_moving(msg->hwnd,(LPRECT)msg->lParam,msg->wParam));
}

#endif


os_error * wmscroll(HWND hWnd,int dx,int dy,CONST RECT *prcScroll,CONST RECT *prcClip,HRGN hrgnUpdate,LPRECT prcUpdate,UINT flags,int * result)
{
 os_error * err;

 drageffectscroll((wimp_w)hWnd,0);
 err=XScrollWindowEx(hWnd,dx,dy,prcScroll,prcClip,hrgnUpdate,prcUpdate,flags,result);
 drageffectscroll((wimp_w)hWnd,1);

 return(err);
} 




static int vscrollf(HWND hwnd,WPARAM wParam)
{
 os_error   * err;
 SCROLLINFO   si;
 RECT         rect;
 int          oldpos;
 int          vpos;
 int          ey0;
 int          ey1;
 int          code;
 int          limit;
 int          size;
 int          event;
 wimp_w       handle;
 int          hit;
 wimp_wstate  winds;
 eventscanstr es;
 int          dsize;


// cprintf("vscrollf");

 err=XGetClientRect(hwnd,&rect);
 size=ypix2os(rect.bottom-rect.top);

 code=LOWORD(wParam);

 XZeroMemory(&si, sizeof(SCROLLINFO));
 si.cbSize=sizeof(SCROLLINFO);
 si.fMask=SIF_POS | SIF_RANGE;
 if(code==SB_THUMBTRACK || code==SB_THUMBPOSITION) si.fMask|=SIF_TRACKPOS;
 err=XGetScrollInfo(hwnd,SB_VERT,&si);
 oldpos=si.nPos;
 ey1=si.nMax;
 ey0=si.nMin;
 if(code==SB_THUMBTRACK || code==SB_THUMBPOSITION) vpos=si.nTrackPos;
 else                                              vpos=oldpos;

 limit=ey1-size;

/*
 if(mfnnclastbuttons==0x1) dsize=-size;
 else  */                  dsize=size;


 switch(code)
 {
	 case SB_LINEDOWN:
		 vpos+=dsize/10;
		 if(vpos>limit) vpos=limit;
		 break;

	 case SB_LINEUP:
		 vpos-=dsize/10;
		 if(vpos<ey0) vpos=ey0;
		 break;

	 case SB_THUMBTRACK:
		 if(vpos>limit) vpos=limit;
		 if(vpos<ey0) vpos=ey0;
		 break;

	 case SB_THUMBPOSITION:
		 if(vpos>limit) vpos=limit;
		 if(vpos<ey0) vpos=ey0;
		 break;

	 case SB_PAGEDOWN:
		 vpos+=dsize;
		 if(vpos>limit) vpos=limit;
		 break;

	 case SB_PAGEUP:
		 vpos-=dsize;
		 if(vpos<ey0) vpos=ey0;
		 break;


    case SB_TOP:
                vpos=ey0;
                break;
               
    case SB_BOTTOM:
                   vpos=limit;
                   break;
 }


 if(oldpos!=vpos)
 {
//		cprintf("v scroll old %d vpos %d",oldpos,vpos);

  handle=(wimp_w)hwnd;

  scrollchecky(hwnd,vpos);

  wimp_get_wind_state(handle,&winds);

//  winds.o.y=-vpos;

  starteventscan(&es);
  hit=0;

  while((event=nextevent(&es,wimp_EOPEN,-1,handle,-1))>=0)
  {
   hit=1;
   report(eventstack[event].fn.open(handle,eventstack[event].icon,&winds.o));
  }

  endeventscan(&es);
 }
 return(0);
} 
 


static int hscrollf(HWND hwnd,WPARAM wParam)
{
 os_error   * err;
 SCROLLINFO   si;
 RECT         rect;
 int          oldpos;
 int          hpos;
 int          ex0;
 int          ex1;
 int          code;
 int          limit;
 int          size;
 int          event;
 wimp_w       handle;
 int          hit;
 wimp_wstate  winds;
 eventscanstr es;


 err=XGetClientRect(hwnd,&rect);
 size=xpix2os(rect.right-rect.left);

// cprintf("hscrollf");

 code=LOWORD(wParam);

 XZeroMemory(&si, sizeof(SCROLLINFO));
 si.cbSize=sizeof(SCROLLINFO);
 si.fMask=SIF_POS | SIF_RANGE;
 if(code==SB_THUMBTRACK || code==SB_THUMBPOSITION) si.fMask|=SIF_TRACKPOS;
 err=XGetScrollInfo(hwnd,SB_HORZ,&si);
 oldpos=si.nPos;
 ex1=si.nMax;
 ex0=si.nMin;
 if(code==SB_THUMBTRACK || code==SB_THUMBPOSITION) hpos=si.nTrackPos;
 else                                              hpos=oldpos;

 limit=ex1-size;


 switch(code)
 {
	 case SB_LINEDOWN:
		 hpos+=size/10;
		 if(hpos>limit) hpos=limit;
		 break;

	 case SB_LINEUP:
		 hpos-=size/10;
		 if(hpos<ex0) hpos=ex0;
		 break;

	 case SB_THUMBTRACK:
		 if(hpos>limit) hpos=limit;
		 if(hpos<ex0) hpos=ex0;
		 break;

	 case SB_THUMBPOSITION:
		 if(hpos>limit) hpos=limit;
		 if(hpos<ex0) hpos=ex0;
		 break;

	 case SB_PAGEDOWN:
		 hpos+=size;
		 if(hpos>limit) hpos=limit;
		 break;

	 case SB_PAGEUP:
		 hpos-=size;
		 if(hpos<ex0) hpos=ex0;
		 break;

  case SB_TOP:
                hpos=ex0;
                break;
               
  case SB_BOTTOM:
                 hpos=limit;
                 break;

 }


 if(oldpos!=hpos)
 {
  handle=(wimp_w)hwnd;

  scrollcheckx(hwnd,hpos);
  
  wimp_get_wind_state(handle,&winds);
//  winds.o.x=hpos;

  starteventscan(&es);
  hit=0;

  while((event=nextevent(&es,wimp_EOPEN,-1,handle,-1))>=0)
  {
   hit=1;
   report(eventstack[event].fn.open(handle,eventstack[event].icon,&winds.o));
  }

  endeventscan(&es);
 }

 return(0);
} 
 




os_error * addopenevent(openfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(wimp_EOPEN,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remopenevent(openfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(wimp_EOPEN,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

/*****************************************************************************/


static int editupdate(HWND hwnd,HWND hwndCtl,int wID)
{
 int          event;
 eventscanstr es;

 starteventscan(&es);

 while((event=nextevent(&es,EDITUPDATE,-1,(wimp_w)hwnd,-1))>=0)
 {
  report(eventstack[event].fn.editupdate((wimp_w)hwnd,wID,eventstack[event].icon));
 }

 endeventscan(&es);

 return(0);

 USE(hwndCtl);
}



os_error * addeditevent(editupdatefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(EDITUPDATE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


os_error * remeditevent(editupdatefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(EDITUPDATE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


/*****************************************************************************/


static int editcaret(HWND hwnd,HWND hwndCtl,int wID,int gain)
{
 int          event;
 eventscanstr es;

 starteventscan(&es);

 while((event=nextevent(&es,EDITCARET,-1,(wimp_w)hwnd,-1))>=0)
 {
  report(eventstack[event].fn.editcaret((wimp_w)hwnd,wID,gain,eventstack[event].icon));
 }

 endeventscan(&es);

 return(0);

 USE(hwndCtl);
}



os_error * addeditcaretevent(editcaretfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(EDITCARET,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


os_error * remeditcaretevent(editcaretfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(EDITCARET,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


/****************************************************************************/


// called on WM_SETFOCUS and WM_KILLFOCUS
// all this is messed up somewhat by icons a window can lose the caret when
// a child icon gains it

static int caretevent(MSG * msg)
{
 int          event;
 wimp_w       handle;
 eventscanstr es;
 int          gain;

 wmfocus(msg->hwnd,msg->message);

 gain=(msg->message==WM_SETFOCUS);
 handle=(wimp_w)msg->hwnd;
 starteventscan(&es);

// cprintf("gain %d handle %x wparam %x",gain,handle,msg->wParam);

 while((event=nextevent(&es,wimp_ELOSECARET,-1,handle,-1))>=0)
 {
  report(eventstack[event].fn.caret(handle,eventstack[event].icon,gain));
 }

 endeventscan(&es);

 return(0);
}

os_error * addcaretevent(caretfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(wimp_ELOSECARET,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remcaretevent(caretfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(wimp_ELOSECARET,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


/****************************************************************************/

// return 1 to let system proceed

/* 
   When the wParam parameter is FALSE, an application should return TRUE
   to indicate that the system should proceed with the default processing,
   or it should return FALSE to prevent the title bar or icon from being
   deactivated. When wParam is TRUE, the return value is ignored
*/


static int wmncactivate(MSG * msg)
{
 int  code;

 code=wimp_ncactivate(msg->hwnd,msg->wParam);


 return(code);
}






static int wmactivate(MSG * msg)
{
 int          event;
 wimp_w       handle;
 eventscanstr es;
 int          code;
 wimp_w       otherhandle;

 handle=(wimp_w)msg->hwnd;
 code=LOWORD(msg->wParam);
 starteventscan(&es);
 otherhandle=(wimp_w)msg->lParam;

// cprintf("wmact code=%d otherhanlde=%d",code,otherhandle);

 while((event=nextevent(&es,ACTIVATE,-1,handle,-1))>=0)
 {
  report(eventstack[event].fn.activate(handle,eventstack[event].icon,otherhandle,code));
 }

 endeventscan(&es);

 return(0);
}


os_error * addactevent(activatefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(ACTIVATE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remactevent(activatefn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(ACTIVATE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}



/****************************************************************************/


static int wmhotkey(MSG * msg)
{
 int          event;
 wimp_w       handle;
 eventscanstr es;

 handle=(wimp_w)msg->hwnd;
 starteventscan(&es);

  cprintf("wm hotkey %d ",msg->wParam);

 while((event=nextevent(&es,HOTKEY,-1,handle,msg->wParam))>=0)
 {
  report(eventstack[event].fn.hot(handle,eventstack[event].icon,eventstack[event].message));
 }

 endeventscan(&es);

 return(0);
}





os_error * addhotkey(hotkeyfn fn,wimp_w handle,uservalue userhandle,int key)
{
 os_error * err;
 int        hit;
 int        modifiers;
 int        vk;
 
 err=addblock(HOTKEY,key,handle,userhandle,(zerofn)fn,&hit);
 if(!err)
 {
  vk=keycanonical2code(key);

//		cprintf("vk %x key %x",vk,key);

//		cprintf("vk %x KEYCTRL %x and %x",vk,KEYCTRL,vk & KEYCTRL);

		modifiers=0;

  if(vk & KEYVSHIFT) modifiers|=MOD_SHIFT;
  if(vk & KEYVCTRL)  modifiers|=MOD_CONTROL;
  if(vk & KEYVALT)   modifiers|=MOD_ALT;
  if(vk & KEYVWIN)   modifiers|=MOD_WIN;

  vk=vk & KEYVMASK;

//		cprintf("reg hot key mod %x vk %x",modifiers,vk);

  if(hit) err=XRegisterHotKey((HWND)handle,userhandle,modifiers,vk);
 }

 return(err);
}

os_error * remhotkey(hotkeyfn fn,wimp_w handle,uservalue userhandle,int key)
{
 os_error * err;
 int        hit;

 err=remblock(HOTKEY,key,handle,userhandle,(zerofn)fn,&hit);
 if(!err)
 {
  if(hit) UnregisterHotKey((HWND)handle,userhandle);

 }
 
 return(err);
}





/****************************************************************************/


os_error * adddataload(wimp_w handle,uservalue userhandle,loadfn fn)
{
 os_error * err;
 err=addblock(DATALOAD,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * adddataloadpost(wimp_w handle,uservalue userhandle,loadpostfn fn)
{
 os_error * err;
 err=addblock(DATALOADPOST,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * adddataloadtype(wimp_w handle,uservalue userhandle,loadtypefn fn)
{
 os_error * err;
 err=addblock(DATALOADTYPE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * adddataloadname(wimp_w handle,uservalue userhandle,loadnamefn fn)
{
 os_error * err;
 err=addblock(DATALOADNAME,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remdataload(wimp_w handle,uservalue userhandle,loadfn fn)
{
 os_error * err;
 err=remblock(DATALOAD,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remdataloadpost(wimp_w handle,uservalue userhandle,loadpostfn fn)
{
 os_error * err;
 err=remblock(DATALOADPOST,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remdataloadtype(wimp_w handle,uservalue userhandle,loadtypefn fn)
{
 os_error * err;
 err=remblock(DATALOADTYPE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remdataloadname(wimp_w handle,uservalue userhandle,loadnamefn fn)
{
 os_error * err;
 err=remblock(DATALOADNAME,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

/*****************************************************************************/




static os_error * dataopens(char * name,int type,int xvolatile,int * ack)
{ 
 os_error   * err;
 int          event;
 eventscanstr es;

 err=NULL;

 if(ack) *ack=0;

 starteventscan(&es);

 if((event=nextevent(&es,DATAOPEN,-1,(wimp_w)type,-1))>=0)
 {
  err=eventstack[event].fn.load(name,type,0,xvolatile);

  if(ack) *ack=1;

  restarteventscan(&es);

  if(!err)
  {
   if((event=nextevent(&es,DATAOPENPOST,-1,(wimp_w)type,-1))>=0)
   {
    err=eventstack[event].fn.loadpost(0);
   }
  }
 }

 endeventscan(&es);

 return(err);

}


typedef struct dataopenstr
{
 int  type;
 char name[1];

} dataopenstr;


static os_error * dataopenevent(int eventn,uservalue userhandle,void * data,int data1)
{
 dataopenstr * dataopendata;

 dataopendata=(dataopenstr*)userhandle;

 report(dataopens(dataopendata->name,dataopendata->type,0,NULL));

 remevent(EVENT_TASKRUN,(eventfn)dataopenevent,(uservalue)dataopendata);
 sfree((flex_ptr)&dataopendata);

 return(NULL);

 USE(eventn);
 USE(data);
 USE(data1);
}


os_error * dataopen(char * name,int xvolatile,int * ack)
{
 int           type;
 int           len;
 dataopenstr * dataopendata;

 type=fs_filetypefromextension(name);

 if(!xvolatile && !taskrunning())
 {
  // there is a problem, error box opens during run, DDE open message comes in whilst error box
  // is open holding things up. Program init not complete.

  *ack=1;
  len=strlen(name);
  salloc((flex_ptr)&dataopendata,sizeof(dataopenstr)+len+1);
  strcpy(dataopendata->name,name);
  dataopendata->type=type;
  addevent(EVENT_TASKRUN,(eventfn)dataopenevent,(uservalue)dataopendata);
  return(NULL);
 }
 else
 {
  return(dataopens(name,type,xvolatile,ack));
 }
}


os_error * fakedataopen(int type,char * name,int * ack)
{
 return(dataopens(name,type,0,ack));
}




os_error * adddataopen(int type,loadfn fn)
{
 os_error * err;
 err=addblock(DATAOPEN,0,(wimp_w)type,0,(zerofn)fn,NULL);
 return(err);
}

os_error * adddataopenpost(int type,loadpostfn fn)
{
 os_error * err;
 err=addblock(DATAOPENPOST,0,(wimp_w)type,0,(zerofn)fn,NULL);
 return(err);
}


os_error * remdataopen(int type,loadfn fn)
{
 os_error * err;
 err=remblock(DATAOPEN,0,(wimp_w)type,0,(zerofn)fn,NULL);
 return(err);
}

os_error * remdataopenpost(int type,loadpostfn fn)
{
 os_error * err;
 err=remblock(DATAOPENPOST,0,(wimp_w)type,0,(zerofn)fn,NULL);
 return(err);
}


/*****************************************************************************/

// return 0 if message handled


#define WINSCANLEVEL0 0
#define WINSCANLEVEL1 1

static int scanwinmessage(int level,MSG * msg,int * code,int * reverselogic)
{
 int          event;
 int          ack;
 int          acked;
 eventscanstr es;

 acked=1;

 starteventscan(&es);

 while((event=nextevent(&es,level==WINSCANLEVEL0?WINMESSAGE0:WINMESSAGE1,msg->message,(wimp_w)msg->hwnd,-1))>=0)
 {
  ack=0;
  if(eventstack[event].fn.message)
  {
   report(eventstack[event].fn.winmessage(msg,eventstack[event].icon,&ack,code,reverselogic));
  }
  if(ack && acked)
  {
   acked=0;
  }
 }

 endeventscan(&es);


 starteventscan(&es);

 while((event=nextevent(&es,level==WINSCANLEVEL0?WINMESSMASK0:WINMESSMASK1,-1,(wimp_w)msg->hwnd,-1))>=0)
 {
  ack=0;
  if(eventstack[event].fn.message)
  {
   report(eventstack[event].fn.winmessage(msg,eventstack[event].icon,&ack,code,reverselogic));
  }
  if(ack && acked)
  {
   acked=0;
  }
 }

 endeventscan(&es);



 starteventscan(&es);

 while((event=nextevent(&es,level==WINSCANLEVEL0?WINMASK0:WINMASK1,-1,(wimp_w)-1,-1))>=0)
 {
  ack=0;
  if(eventstack[event].fn.message)
  {
   report(eventstack[event].fn.winmessage(msg,eventstack[event].icon,&ack,code,reverselogic));
  }
  if(ack && acked)
  {
   acked=0;
  }
 }

 endeventscan(&es);
 
 
 
 return(acked);
}



os_error * addwinmessage(winmessfn fn,int message,wimp_w handle,uservalue userhandle)
{
 os_error * err;

 if(message==-1 && handle==(wimp_w)-1) err=addblock(WINMASK1,message,handle,userhandle,(zerofn)fn,NULL);
	else
	if(message==-1)																							err=addblock(WINMESSMASK1,message,handle,userhandle,(zerofn)fn,NULL);
 else                                  err=addblock(WINMESSAGE1,message,handle,userhandle,(zerofn)fn,NULL);

 return(err);
}


os_error * remwinmessage(winmessfn fn,int message,wimp_w handle,uservalue userhandle)
{
 os_error * err;

 if(message==-1 && handle==(wimp_w)-1) err=remblock(WINMASK1,message,handle,userhandle,(zerofn)fn,NULL);
	else
	if(message==-1)																						 err=remblock(WINMESSMASK1,message,handle,userhandle,(zerofn)fn,NULL);
 else                                  err=remblock(WINMESSAGE1,message,handle,userhandle,(zerofn)fn,NULL);

 return(err);
}


// earlier windows message intercept, needed for TWAIN


os_error * addwinmessage0(winmessfn fn,int message,wimp_w handle,uservalue userhandle)
{
 os_error * err;

 if(message==-1 && handle==(wimp_w)-1) err=addblock(WINMASK0,message,handle,userhandle,(zerofn)fn,NULL);
 else                                  err=addblock(WINMESSAGE0,message,handle,userhandle,(zerofn)fn,NULL);

 return(err);
}


os_error * remwinmessage0(winmessfn fn,int message,wimp_w handle,uservalue userhandle)
{
 os_error * err;

 if(message==-1 && handle==(wimp_w)-1) err=remblock(WINMASK0,message,handle,userhandle,(zerofn)fn,NULL);
 else                                  err=remblock(WINMESSAGE0,message,handle,userhandle,(zerofn)fn,NULL);

 return(err);
}



/*****************************************************************************/


// return 0 if message handled

static int scanmessage(wimp_msgstr * msg)
{
 int          event;
 int          ack;
 int          acked;
 eventscanstr es;

 acked=1;

 starteventscan(&es);

 while((event=nextevent(&es,wimp_ESEND,msg->hdr.action,(wimp_w)-1,-1))>=0)
 {
//  cprintf("event=%d eventn=%d eventstack=%x",event,eventn,eventstack);
  ack=0;
  if(eventstack[event].fn.message)
                          report(eventstack[event].fn.message(msg,&ack));
  if(ack && acked)
  {
   acked=0;
  }
 }

 endeventscan(&es);

 return(acked);
}


int wmgeneratemessage2(wimp_msgstr * msg)
{
 return(scanmessage(msg));
}




int wmgeneratemessage(wimp_msgaction action)
{
 wimp_msgstr xmsg;

 xmsg.hdr.action=action;
 scanmessage(&xmsg);
 return(0);
}


static int wmfontchange(MSG * msg) 
{
 return(wmgeneratemessage(wimp_MFONTCHANGE));
 USE(msg);
}



static void terminate(MSG * msg) 
{
 wmgeneratemessage(wimp_MCLOSEDOWN);
 exit(0);
 USE(msg);
}



/* set up mode specific things like width */

static void setmodevars(MSG * msg)
{
 vdumodevars();

 wmgeneratemessage(wimp_PALETTECHANGE);
 USE(msg);
}




/* set up palette */

static void setpalvars(MSG * msg)
{
 int         code;
 
 palettechanged((HWND)msg->wParam,msg->hwnd,&code);

 wmgeneratemessage(wimp_PALETTECHANGE);
 USE(msg);
}





static int querypal(HWND hwnd)
{
 int code;

 palettequery(hwnd,&code);

 return(code);
}




/****************************************************************************/

#ifdef NEVER

os_error * addprequit(prequitfn fn,uservalue userhandle)
{
 os_error * err;
 err=addblock(PREQUIT,0,userhandle,0,(zerofn)fn,NULL);
 return(err);
}

os_error * remprequit(prequitfn fn,uservalue userhandle)
{
 os_error * err;
 err=remblock(PREQUIT,0,userhandle,0,(zerofn)fn,NULL);
 return(err);
}


#ifdef NEVER

static void prequit(wimp_msgstr * msg)
{
 int  event;
 int  ack;
 int  acked;
 eventscanstr es;


 starteventscan(&es);

 acked=0;

 while((event=nextevent(&es,PREQUIT,-1,-1,-1))>=0)
 {
  ack=0;
  report(eventstack[event].fn.prequit(eventstack[event].handle,&ack,1));
  acked|=ack;
 }

 endeventscan(&es);


 if(acked)   /* stop the prequit sequence */
 {


  starteventscan(&es);

  acked=0;
  while((event=nextevent(&es,PREQUIT,-1,-1,-1))>=0)
  {
   ack=0;
   report(eventstack[event].fn.prequit(eventstack[event].handle,&ack,0));
   acked|=ack;
  }

  endeventscan(&es);

  poll();

  if(!acked)
  {


  }
 }
}


#endif

#endif

/*****************************************************************************/

os_error * helpicon(wimp_w handle,int icon,wimp_w iconhandle)
{
 os_error     * err;
 char         * p;
 eventscanstr   es;
 int            event;
 POINT          pt;

//  cprintf("control help %x id=%d",((LPHELPINFO)lParam)->hItemHandle,((LPHELPINFO)lParam)->iCtrlId);

// cprintf("helpicon handle %x icon %d iconhandle %x",handle,icon,iconhandle);

 err=NULL;

 starteventscan(&es);

 p=NULL;

 if((event=nextevent(&es,HELPMESSAGE,-1,handle,-1))>=0)
 {
  if(eventstack[event].fn.help)
  {
   err=eventstack[event].fn.help(handle,eventstack[event].icon,icon,&p);
  }
  else
  {
   p=helpfindtagstring(eventstack[event].icon,icon);
//   cprintf("help find %d %d %s",eventstack[event].icon,icon,p);
  }
 }

 endeventscan(&es);
 
 if(p)
 {
  pt.x=pt.y=0;
  err=XClientToScreen(iconhandle?((HWND)iconhandle):((HWND)handle),&pt);

  helppopup(handle,p,pt.x,pt.y);
 }

 return(err);
}





static int helpcommand(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam)
{

 if(GetKeyState(VK_F1) & 0x8000)
 {
  if(GetKeyState(VK_CONTROL) & 0x8000) return(0);
  if(GetKeyState(VK_MENU) & 0x8000) return(0);
  if(GetKeyState(VK_SHIFT) & 0x8000)
  {
   helpwhatsthis((wimp_w)hwnd);
   return(0);
  }
 }

 if(((LPHELPINFO)lParam)->iContextType==HELPINFO_MENUITEM)
 {
//  cprintf("menu help %x",((LPHELPINFO)lParam)->iCtrlId);
  menuhelp(hwnd,((LPHELPINFO)lParam)->iCtrlId,(HMENU)((LPHELPINFO)lParam)->hItemHandle);
//  XWinHelp(hwnd,NULL,HELP_CONTEXTPOPUP,(DWORD)(((LPHELPINFO)lParam)->iCtrlId+0x10000));
 }
 else
 {
  if(((LPHELPINFO)lParam)->hItemHandle!=hwnd)
  {
   helpicon((wimp_w)hwnd,((LPHELPINFO)lParam)->iCtrlId,(wimp_w)((LPHELPINFO)lParam)->hItemHandle);
  }
  else
  {
 //  XWinHelp(hwnd,NULL,HELP_KEY,(DWORD)"Main Window");
 //  cprintf("other help");
   helpicon((wimp_w)hwnd,-1,0);
  }
 }

 return(0);
 USE(wParam);
 USE(message);
}










// right click - can be on window title, in which case system menu should appear
// in fact there will be no client events because the mouse code will
// consume all right clicks, anything you get is going to be non-client

static int contextmenu(MSG * msg)
{
 HWND child;
 int  id;
 int  hitcode;

 child=(HWND)msg->wParam;

// cprintf("context menu");

 if(child==msg->hwnd)
 {
  wimp_mouseitem(msg->hwnd,GET_X_LPARAM(msg->lParam),GET_Y_LPARAM(msg->lParam),&hitcode);
  if(hitcode==HTCLIENT)
  {
   helpicon((wimp_w)child,-1,0);
   return(0);
  }
 }
 else
 {
  XGetDlgCtrlID(child,&id);
  helpicon((wimp_w)msg->hwnd,id,(wimp_w)child);
  return(0);
 }

 return(1); // pass on
}




os_error * addhelpevent(helpfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(HELPMESSAGE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remhelpevent(helpfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(HELPMESSAGE,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


/*****************************************************************************/



os_error * addmessage(messfn fn,int message)
{
 os_error * err;
 err=addblock(wimp_ESEND,message,0,0,(zerofn)fn,NULL);
 return(err);
}

os_error * remmessage(messfn fn,int message)
{
 os_error * err;
 err=remblock(wimp_ESEND,message,0,0,(zerofn)fn,NULL);
 return(err);
}


/****************************************************************************/

#ifdef NEVER

static ackfn ack;
static int   ackref;

os_error * addack(ackfn fn,int ref)
{
 ack=fn;
 ackref=ref;
 return(NULL);
}




static os_error * acknowledged(MSG * msg)
{
 return(ack(msg));
}

#endif


/***************************************************************************/

#define wimp_POLLWORD 13
#define POLLWORDMASK (1<<22)


static int * pollword;

static pollwordfn pollwordhandler;


os_error * addpollword(pollwordfn fn,int * pw)
{
 pollword=pw;
 pollwordhandler=fn;
 return(NULL);
}

os_error * rempollword(void)
{
 pollword=NULL;
 return(NULL);
}


#ifdef NEVER

static void pollwordevent(void)
{

}

#endif


/**************************************************************************/

static int wmcommand(HWND hwnd,WPARAM wParam,LPARAM lParam)
{
 int  wNotifyCode;
 int  wID;
 HWND hwndCtl;
 int  code;
 MSG  msg;

 code=1;
 wNotifyCode=HIWORD(wParam); // notification code 
 wID=LOWORD(wParam);         // item, control, or accelerator identifier 
 hwndCtl=(HWND)lParam;       // handle of control 

// cprintf("notify %d id=%d hwndctl %x",wNotifyCode,wID,hwndCtl);

 if(hwndCtl)  // we know this is a control that has been clicked
 {
  if(wNotifyCode==BN_CLICKED) code=mousecommand(hwnd,hwndCtl,wID);
  else
  if(wNotifyCode==EN_UPDATE) code=editupdate(hwnd,hwndCtl,wID);
  else
  if(wNotifyCode==EN_KILLFOCUS) code=editcaret(hwnd,hwndCtl,wID,0);
  else
  if(wNotifyCode==EN_SETFOCUS) code=editcaret(hwnd,hwndCtl,wID,1);
 }
 else
 {
  // wNotifyCode
  // If the message is from an accelerator, this parameter is 1. 
  // If the message is from a menu, this parameter is 0. 
  // Return in an Edit control will fake a click on IDOK - if there isn't one we end up here
  // and wNotifyCode will be 0...

  if(wID==IDOK || wID==IDCANCEL)
  {
   msg.message=WM_CHAR;
   msg.hwnd=hwnd;
   msg.lParam=0;
   msg.wParam=(wID==IDOK)?13:27;
   code=keypress(&msg,0);
  }
  else
  {
   if(wNotifyCode==0) code=menucommand(hwnd,wParam,lParam);
  }
 }

 return(code);
}


/*******************************************************************************/



int dataload(char * name,wimp_w loadhandle,mousestr * mouse)
{
 os_error   * err;
 int          event;
 eventscanstr es;
 int          type;
 int          code;
 int          load;
 fstat        f;


 err=NULL;

 load=1;
 type=fs_filetypefromextension(name);

 fs_stat(name,&f);
 if(f.object==2) type+=DIR;

 starteventscan(&es);
 if((event=nextevent(&es,DATALOADTYPE,-1,loadhandle,-1))>=0)
 {
  code=0;
  err=eventstack[event].fn.loadtype(name,type,mouse,eventstack[event].icon,&code);
  if(!err && code)
  {
   load=1;
   restarteventscan(&es);
   if((event=nextevent(&es,DATALOAD,-1,loadhandle,-1))>=0)
   {
    err=eventstack[event].fn.load(name,type,eventstack[event].icon,0);
    if(!err)
    {
     restarteventscan(&es);
     if((event=nextevent(&es,DATALOADPOST,-1,loadhandle,-1))>=0)
     {
      err=eventstack[event].fn.loadpost(eventstack[event].icon);
     }
    }
   }
  }
  report(err);
 }
 endeventscan(&es);

 return(load);
}



static int ddfile(HWND hwnd,WPARAM wParam)
{
 os_error   * err;


 HDROP        hdrop;
 POINT        pt;
 mousestr     mouse;
 unsigned int nofiles;
 unsigned int i;

 char         name[MAX_PATH];
 wimp_w       loadhandle;

 HWND         child;
 unsigned int count;
 int          result;


 err=NULL;

 hdrop=(HDROP)wParam;


 loadhandle=(wimp_w)hwnd;

// cprintf("ddfile");
 
 XDragQueryPoint(hdrop,&pt,&result);
 XChildWindowFromPoint(hwnd,pt,&child);
 if(child)
 {
  XGetDlgCtrlID(child,&mouse.icon);
 }
 else
 {
  mouse.icon=-1;
 }

 XClientToScreen(hwnd,&pt);
 mouse.x=xcoordpix2os(pt.x);
 mouse.y=ycoordpix2os(pt.y);
 mouse.handle=(wimp_w)hwnd;

// cprintf("ddfiles %d",(int)hwnd);

 XDragQueryFile(hdrop,0xFFFFFFFF,name,MAX_PATH,&nofiles);

 for(i=0;i<nofiles;i++)
 {
  XDragQueryFile(hdrop,i,name,MAX_PATH,&count);
  dataload(name,loadhandle,&mouse);
 }

 XDragFinish(hdrop);


 return(0);
}





/******************************************************************************/

// return 1 if we do something

static int systemcommand(MSG * msg)
{

 return(wimp_systemcommand(msg->hwnd,msg->wParam,
                            GET_X_LPARAM(msg->lParam),GET_Y_LPARAM(msg->lParam)));    
}


/******************************************************************************/



static int wmappcommand(MSG * msg)
{
 int          event;
 eventscanstr es;
	int          hit;
 wimp_w       handle;
	wimp_w       source;
 int          cmd;
 int          device;
 int          buttons;


 hit=0;
	handle=(wimp_w)msg->hwnd;
 source=(wimp_w)msg->wParam;
 cmd=GET_APPCOMMAND_LPARAM(msg->lParam);
 device=GET_DEVICE_LPARAM(msg->lParam);
 buttons=GET_KEYSTATE_LPARAM(msg->lParam);

 starteventscan(&es);

 while((event=nextevent(&es,APPCOMMAND,-1,handle,-1))>=0)
 {
  report(eventstack[event].fn.appcommand(handle,eventstack[event].icon,source,cmd,device,buttons,&hit));
  break;
 }

 endeventscan(&es);


	return(hit!=0); // return true if we process this
}




os_error * addappcommandevent(appcommandfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(APPCOMMAND,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}


os_error * remappcommandevent(appcommandfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(APPCOMMAND,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}



/*******************************************************************************/


static void wheelevent(wimp_w handle,int delta,int buttons,int horizontal,int * hit)
{
 int          event;
 eventscanstr es;

 *hit=0;

 starteventscan(&es);

 while((event=nextevent(&es,WHEEL,-1,handle,-1))>=0)
 {
  report(eventstack[event].fn.wheel(handle,eventstack[event].icon,delta,buttons,horizontal,hit));
  break;
 }

 endeventscan(&es);
}



static int wheelwindowsub(wimp_w handle,int delta,int buttons,int horizontal)
{
 int code;

 wheelevent(handle,delta,buttons,horizontal,&code);
 if(!code)
 {
  if(horizontal) code=wheelhwindow(HWND(handle),delta,buttons);
  else           code=wheelwindow(HWND(handle),delta,buttons);
 }
 else
 {
  if(!horizontal) code=!code;
 }

 return(code);
}



os_error * addwheelevent(wheelfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=addblock(WHEEL,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}

os_error * remwheelevent(wheelfn fn,wimp_w handle,uservalue userhandle)
{
 os_error * err;
 err=remblock(WHEEL,0,handle,userhandle,(zerofn)fn,NULL);
 return(err);
}







static int wmwheel(MSG * msg)
{
 int   code;
 POINT pt;
 HWND  hwnd;
 HWND  child;

 code=wheelwindowsub((wimp_w)msg->hwnd,((short)HIWORD(msg->wParam)),((short)LOWORD(msg->wParam)),0);
 if(code)
 {
  hwnd=msg->hwnd;
  pt.x=(short)LOWORD(msg->lParam);
  pt.y=(short)HIWORD(msg->lParam);    
  ScreenToClient(hwnd,&pt);

  while(1)
  {
   child=ChildWindowFromPoint(hwnd,pt);
/*
   if(child && child!=hwnd) cprintf("try child %x",child);
   if(child==NULL) cprintf("outside window");
   else
   if(child==hwnd) cprintf("outside child");
*/
   if(child && child!=hwnd) hwnd=child;
   else                     break;

// check child is task class
   if(!taskclasswindow(child)) break;

   code=wheelwindowsub((wimp_w)child,((short)HIWORD(msg->wParam)),((short)LOWORD(msg->wParam)),0);
   if(!code) break;
  }
 }
 return(code);
}



static int wmhwheel(MSG * msg)
{
 int   code;
 POINT pt;
 HWND  hwnd;
 HWND  child;

 code=wheelwindowsub((wimp_w)msg->hwnd,((short)HIWORD(msg->wParam)),((short)LOWORD(msg->wParam)),1);
 if(!code)
 {
  hwnd=msg->hwnd;
  pt.x=(short)LOWORD(msg->lParam);
  pt.y=(short)HIWORD(msg->lParam);    
  ScreenToClient(hwnd,&pt);

  while(1)
  {
   child=ChildWindowFromPoint(hwnd,pt);
/*
   if(child && child!=hwnd) cprintf("try child %x",child);
   if(child==NULL) cprintf("outside window");
   else
   if(child==hwnd) cprintf("outside child");
*/
   if(child && child!=hwnd) hwnd=child;
   else                     break;

// check child is task class
   if(!taskclasswindow(child)) break;

   code=wheelwindowsub((wimp_w)child,((short)HIWORD(msg->wParam)),((short)LOWORD(msg->wParam)),1);
   if(code) break;
  }
 }
 return(code);
}



/******************************************************************************/



static int wmsettingchange(MSG * msg)
{

 osgluechange(msg->wParam,msg->lParam);


 return(0);
}



/******************************************************************************/

static int wmmeasureitem(MSG * msg)
{

 return(0);
 USE(msg);

}

static int wmdrawitem(MSG * msg)
{
 int code;

 code=drawitem(msg->hwnd,(UINT)msg->wParam,(LPDRAWITEMSTRUCT)msg->lParam);

 return(code);
}



// worry do we need to close font handle on delete window?

static int wmsetfont(HWND hwnd,HFONT font)
{
// cprintf("wmset font %x",font);
 
 SetWindowLongPtr(hwnd,DWLP_USER,(LONG_PTR)font);
 return(0);
}



/******************************************************************************/


LRESULT CALLBACK pollevent(HWND hwnd, UINT message, 
                           WPARAM wParam, LPARAM lParam)
{
 MSG msg;
 int code;
 int reverselogic;

// static int pollcount;

// pollcount++;
// cprintf("pc %d",pollcount);

 code=1;
 reverselogic=0;

 msg.hwnd=hwnd;
 msg.lParam=lParam;
 msg.message=message;
 msg.wParam=wParam;

 threadlevel++;

 if(scanwinmessage(WINSCANLEVEL1,&msg,&code,&reverselogic))
 {
  switch(message)
  {

            case WM_COMMAND:
                            code=wmcommand(hwnd,wParam,lParam);
    		                      break;

          case WM_DROPFILES:
                            code=ddfile(hwnd,wParam);
                            break;


              case WM_PAINT:
			                         code=redraw(&msg);
			                         break;

												case	WM_IME_CHAR:
																													cprintf("******************* IME char %x",msg.wParam);
																												code=keypress(&msg,0);
																												break;

										  case	WM_UNICHAR:
																												cprintf("******************* unichar");


            case WM_KEYDOWN:
         case WM_SYSKEYDOWN:
               case WM_CHAR:
                            code=keypress(&msg,0);
	                           break;
 
            case WM_DESTROY:
                            code=0;
	                           // PostQuitMessage(0);
	                           break;

              case WM_CLOSE:
                            code=wclose(&msg);
                            break;


               case WM_HELP:
                             cprintf("wm help");
                            code=helpcommand(hwnd,message,wParam,lParam);
                            break;

        case WM_CONTEXTMENU:
                            // cprintf("context menu");
                            code=contextmenu(&msg);
                            break;


//         case WM_GETDLGCODE:
//                          return(DLGC_DEFPUSHBUTTON);
//                            break;


#ifdef NEVER
          case WM_GETDLGCODE:
                            cprintf("get dlg code");
                            // DLGC_BUTTON - means can't be checked
                            // need DEFPUSHBUTTON to get returns which get passed as WM_COMMAND's to the parent
                            return(DLGC_ALLKEYS);
                            return(DLGC_WANTCHARS);  // prob need to return Radio for some
                            break;
#endif





//    case WM_NCLBUTTONDBLCLK:
//      case WM_NCLBUTTONDOWN:
//        case WM_NCLBUTTONUP:
//       case WM_NCMOUSEMOVE:
//    case WM_NCRBUTTONDBLCLK:
//      case WM_NCRBUTTONDOWN:
//        case WM_NCRBUTTONUP:
//                            mouseevent(hwnd);
//          case WM_NCHITTEST:
//                            cprintf("NC mouse");

//                            break;


      case WM_LBUTTONDBLCLK:
        case WM_LBUTTONDOWN:
          case WM_LBUTTONUP:
          case WM_MOUSEMOVE:
  
      case WM_RBUTTONDBLCLK:
        case WM_RBUTTONDOWN:
          case WM_RBUTTONUP:
//                            if(message!=WM_MOUSEMOVE) mouseevent(hwnd);
                            code=mousefn(hwnd,message,wParam,lParam);
//                            if(message!=WM_MOUSEMOVE) cprintf("mouse code=%d",code);
                            break;

            case WM_VSCROLL:
                            code=vscrollf(hwnd,wParam);
                            break;

            case WM_HSCROLL:
                            code=hscrollf(hwnd,wParam);
                            break;

         case WM_MENUSELECT:
                            code=menuselect(hwnd,wParam,lParam);
                            break;

      case WM_INITMENUPOPUP:
	                           code=menuinitpopup(hwnd,wParam,lParam);
                            break;

           case WM_INITMENU:
                            code=menuinit(hwnd,wParam,lParam);
                            break;

               case WM_MOVE:
               case WM_SIZE:
         case WM_SHOWWINDOW:

/*          
                            if(message==WM_MOVE) cprintf("MOVE");
                            else
                            if(message==WM_SIZE) cprintf("SIZE");
                            else
                            if(message==WM_SHOWWINDOW) cprintf("SHOW");
*/           
                            code=wopen(&msg);
                            break;
        

             case WM_MOVING:
                            //cprintf("moving");
                            //reverselogic=1;
                            //code=wmmoving(&msg);
                            break;


             case WM_SIZING:
                            //cprintf("sizing");
                            //reverselogic=1;
                            //code=sizewindow(&msg);
                            break;


      case WM_GETMINMAXINFO:
                            code=wmgetminmax(&msg);
                            break;

           case WM_SETFOCUS:
          case WM_KILLFOCUS:
                            code=caretevent(&msg);
                            break;


           case WM_ACTIVATE:
                            code=wmactivate(&msg);
                            break;

           case   WM_HOTKEY:
                            code=wmhotkey(&msg);
                            break;


         case WM_NCACTIVATE:
                            reverselogic=1;
                            code=wmncactivate(&msg);
                            break;

     case WM_PALETTECHANGED:
                            setpalvars(&msg);
                            break;

    case WM_QUERYNEWPALETTE:
                            reverselogic=1;
                            code=querypal(hwnd);
                            break;

     case WM_SYSCOLORCHANGE:
                            syscolourchange(hwnd);
                            break;

      case WM_DISPLAYCHANGE:
                            setmodevars(&msg);
                            break;

         case WM_SYSCOMMAND:
                            reverselogic=1;
                            //cprintf("syscommand %x",wParam);
                            code=systemcommand(&msg);
                            break;

     case WM_CAPTURECHANGED:
                            code=wmcapture(&msg);
                            break;


         case WM_MOUSEWHEEL:
                            code=wmwheel(&msg);
                            break;


        case WM_MOUSEHWHEEL:
                            reverselogic=1;
                            code=wmhwheel(&msg);
                            break;


      case WM_SETTINGCHANGE:           // could be interesting
                            code=wmsettingchange(&msg);
                            break;


         case WM_APPCOMMAND:
                            cprintf("app command");
																												reverselogic=1;
																												code=wmappcommand(&msg);
                            break;

           case WM_DRAWITEM:
                            code=wmdrawitem(&msg);
                            break;

        case WM_MEASUREITEM:
                            code=wmmeasureitem(&msg);
                            break;

            case WM_SETFONT:
                            code=wmsetfont(hwnd,(HFONT)wParam);
                            break;

         case WM_FONTCHANGE:
                            code=wmfontchange(&msg);
                            break;

         case WM_ERASEBKGND:
                            reverselogic=1;
                            code=windoweraseback(hwnd,(HDC)wParam);
                            break;

          case WM_SETCURSOR:
                            reverselogic=1;
                            hourglasssetcursor(&code);
                            if(!code)
                            {
                             code=wmsetcursor(&msg);
                             // let other code have a go
                            }
                            break;


       case WM_DDE_INITIATE:
                            code=ddeinit(hwnd,(HWND)wParam,LOWORD(lParam),HIWORD(lParam));
                            break;

           case WM_DDE_DATA:
                            cprintf("dde data");
                            code=ddedata(hwnd,(HWND)wParam,lParam);
                            break;

           case WM_DDE_POKE:
                            cprintf("dde poke");
                            code=ddepoke(hwnd,(HWND)wParam,lParam);
                            break;

        case WM_DDE_REQUEST:
                            cprintf("dde req");
                            code=dderequest(hwnd,(HWND)wParam,lParam);
                            break;

      case WM_DDE_TERMINATE:
                            code=ddeterm(hwnd,(HWND)wParam,lParam);
                            break;

        case WM_DDE_EXECUTE:
                            code=ddeexec(hwnd,(HWND)wParam,(HGLOBAL)lParam);
                            break;

            case WM_DDE_ACK:
                            code=ddeack(hwnd,(HWND)wParam,lParam);
                            break;



       case WM_RENDERFORMAT:
                         code=oscbrender(hwnd,wParam);
                         break;

       case WM_RENDERALLFORMATS:
                         code=oscbrenderall(hwnd);
                         break;

       case WM_DESTROYCLIPBOARD:
                         code=oscbdestroy(hwnd);
                         break;

       case           WM_CREATE:
                                code=wmcreate(hwnd);
                                break;

                    default:
//                             cprintf("mess %x",message);
                            break;

  }
 }

 // normal logic commands return 0 if they do something

// cprintf("event message %d code=%d",message,code);

	// code==0 and normal logic don't call def proc
	// code==1 and normal logic	call def proc
	// code==0 and reverse logic call def proc
	// code==1 and reverse logic don't call def proc

 if((code==0)!=(reverselogic==0)) XDefWindowProcX(hwnd,message,wParam,lParam,(LRESULT*)&code);

// pollcount--;

 threadlevel--;

 return(code);
}

/**********************************************************************************/
// SystemParameters to get repeat delay


#define ABLEFT  1
#define ABRIGHT 2

// result of a double click
#define AB2     4

// result of a single click
#define AB1     8
#define ABPUSH  16

// button can send a drag

#define ABREPORTDRAG           32
#define ABREPORTSINGLECLICK    64
#define ABREPORTDOUBLECLICK   128
#define ABREPORTSINGLERELEASE 256
#define ABREPORTDOUBLERELEASE 512

#define ABAUTOREPEAT          1024


#define ABLEFTPUSH1  (ABLEFT|ABPUSH|AB1)
#define ABLEFTPUSH2  (ABLEFT|ABPUSH|AB2)
#define ABRIGHTPUSH1 (ABRIGHT|ABPUSH|AB1)
#define ABRIGHTPUSH2 (ABRIGHT|ABPUSH|AB2)


// pushed and over state are exclusive states


static int  abxstate;
static wimp_w abxhandle;


static void abxmouse(HWND hwnd,int state)
{
 int word;

 word=GetWindowLong(hwnd,ABXFLAGS);
 if(state & ABPUSH) word|=USDPUSHED;
 else               word&=~USDPUSHED;

 
 SetWindowLong(hwnd,ABXFLAGS,word);

 InvalidateRect(hwnd,NULL,TRUE);
 UpdateWindow(hwnd);
}



static int abxcheck(HWND hwnd,WPARAM wParam)
{
 int word;


 word=GetWindowLong(hwnd,ABXFLAGS);
 if(wParam==BST_CHECKED) word|=USDSELECTED;
 else                    word&=~USDSELECTED;
 
 SetWindowLong(hwnd,ABXFLAGS,word);

 InvalidateRect(hwnd,NULL,TRUE);
 UpdateWindow(hwnd);

 return(0);
}



static int abxenable(HWND hwnd,WPARAM wParam)
{
 int word;

 word=GetWindowLong(hwnd,ABXFLAGS);
 if(!wParam) word|=USDGREYED;
 else        word&=~USDGREYED;
 
 SetWindowLong(hwnd,ABXFLAGS,word);

 InvalidateRect(hwnd,NULL,TRUE);
 UpdateWindow(hwnd);

 return(0);
}





// sort out auto repeat

static UINT_PTR     abxtimer=0;
static int          abxcountdown;

static VOID CALLBACK abxproc(HWND hwnd,UINT uMsg,UINT_PTR idEvent,DWORD dwTime)
{
 wimp_mousestr m;

 if(abxcountdown<=0)
 {
  wimp_get_point_info(&m);
  if((wimp_w)hwnd==m.w && idEvent==(unsigned int)m.i)
  {
   m.bbits=(wimp_bbits)((abxstate & ABLEFT)?0x4:0x1);
   decodeclick(&m);
  }                           
 }
 else
 {
  abxcountdown--;
 }
 
 USE(hwnd);
 USE(uMsg);
 USE(idEvent);
 USE(dwTime);
}





static void abxautocancel(HWND hwnd)
{
 if(abxtimer)
 {
  KillTimer(hwnd,abxtimer);
  abxtimer=0;
 }

}

static void abxauto(wimp_mousestr * m,int abxstate)
{

 abxautocancel((HWND)m->w);

 if((abxstate & ABAUTOREPEAT) && (abxstate & (AB1|AB2)))
 {
  abxtimer=SetTimer((HWND)m->w,m->i,100,abxproc);
  abxcountdown=5;

  if(abxstate & ABLEFT) m->bbits=(wimp_bbits)0x4;
  else                  m->bbits=(wimp_bbits)0x1;

  decodeclick(m);
 }
}




// losing capture

static int abxcapture(HWND child,HWND newhwnd)
{
 HWND hwnd;
 int  id;

// cprintf("abxcapture");

 if(abxstate & (AB1|AB2|ABLEFT|ABRIGHT|ABPUSH))
 {
  XGetParent(child,&hwnd);
  XGetDlgCtrlID(child,&id);

  abxstate&=~(AB1|AB2|ABLEFT|ABRIGHT|ABPUSH);
  abxmouse(child,abxstate);
  abxautocancel(hwnd);
 }

 return(0);

 USE(newhwnd);
}


// get button info, set up bits

static void abxsetstate(HWND hwnd)
{
 char   buffer[1024];
 char * p;
 int    ch;
 int    release=0;

 abxread(hwnd,buffer,sizeof(buffer)/sizeof(TCHAR));

 if(getgivenitem(buffer,'K',buffer))
 {
  p=buffer;
  while((ch=*p++)!=0)
  {
   switch(ch)
   {
    case 'D':
             abxstate|=ABREPORTDRAG;
             break;

    case '1':
             abxstate|=release?ABREPORTSINGLERELEASE:ABREPORTSINGLECLICK;
             break;

    case '2':
             abxstate|=release?ABREPORTDOUBLERELEASE:ABREPORTDOUBLECLICK;
             break;

    case 'R':
             release=1;
             break;

    case 'A':
             abxstate|=ABAUTOREPEAT;
             break;

   }
  }
 }
}


// snag here is that all buttons grab mouse clicks


static int abxmousefn(HWND child, UINT message, WPARAM wParam, LPARAM lParam)
{
 int           xpos;
 int           ypos;
 RECT          rect;
 wimp_mousestr m;
 POINT         pt;
 POINT         pt2;
 HWND          hwnd;
 HWND          hprev;
 int           oldstate;
 int           inside;


 pt.x=xpos=(short)LOWORD(lParam);
 pt.y=ypos=(short)HIWORD(lParam);
 pt2=pt;

 XGetParent(child,&hwnd);

 XGetClientRect(child,&rect);
 XClientToScreen(child,&pt2);

 m.x=xcoordpix2os(pt2.x);
 m.y=ycoordpix2os(pt2.y);
 m.w=(wimp_w)hwnd;

 XGetDlgCtrlID(child,&m.i);

 inside=(pt.x>=0 && pt.x<(rect.right-rect.left) && pt.y>=0 && pt.y<(rect.bottom-rect.top));

 switch(message)
 {
 
     case WM_LBUTTONDBLCLK:// like a button down
                           wimp_checkfocus(hwnd);
					                      abxstate|=ABLEFTPUSH2;
                           abxmouse(child,abxstate);
                           XSetCapture(child,&hprev); // has to be here because the click may launch another capture
                           if(abxstate & ABREPORTDOUBLECLICK)
                           {
                            m.bbits=(wimp_bbits)0x400;
                            decodeclick(&m);
                            abxstate&=~ABREPORTDOUBLECLICK;
                           }
                           abxauto(&m,abxstate);
					                      break;

       case WM_LBUTTONDOWN:
                           wimp_checkfocus(hwnd);
                           abxhandle=(wimp_w)hwnd;
					                      abxstate=ABLEFTPUSH1;
                           abxsetstate(child);
                           //cprintf("state1 %x",abxstate);
                           abxmouse(child,abxstate);
                           XSetCapture(child,&hprev);  // has to be here because the click may launch another capture
                           if(abxstate & ABREPORTSINGLECLICK)
                           {
                            m.bbits=(wimp_bbits)0x4;
					                       decodeclick(&m);
                            abxstate&=~ABREPORTSINGLECLICK;
                           }
                           //cprintf("state2 %x",abxstate);
                           abxauto(&m,abxstate);
                           //cprintf("state3 %x",abxstate);
					                      break;

         case WM_LBUTTONUP:
                           //cprintf("l up");
                           oldstate=abxstate;
                           XReleaseCapture();  // has to be here because the click may launch another capture

                           if((oldstate & ABREPORTSINGLERELEASE) && inside)
                           {
                            m.bbits=(wimp_bbits)0x4;
	 					                     decodeclick(&m);
                            abxstate&=~ABREPORTSINGLERELEASE;
                           }

                           if((oldstate & ABREPORTDOUBLERELEASE) &&
                              (oldstate & AB2) && inside)
                           {
                            m.bbits=(wimp_bbits)0x400;
	 					                     decodeclick(&m);
                            abxstate&=~ABREPORTDOUBLERELEASE;
                           }

                           oldstate=abxstate;
 				                      abxstate&=~(AB1|AB2|ABLEFT|ABPUSH|ABREPORTDRAG);
                           //cprintf("clear drag flag");
                           if(oldstate!=abxstate)
                           {
                            abxmouse(child,abxstate);
                           }
                           abxauto(&m,abxstate);
					                      break;
 
         case WM_MOUSEMOVE://cprintf("abx move %x",abxstate);
                           //cprintf("left %x",GetKeyState(VK_LBUTTON) & 0x8000);
                           //cprintf("right %x",GetKeyState(VK_RBUTTON) & 0x8000);
                           
                           if(abxstate & ABREPORTDRAG)
                           {
					                       if(abxstate & ABLEFT) m.bbits=(wimp_bbits)0x40;
                            else                  m.bbits=(wimp_bbits)0x10;

                            abxstate&=~ABREPORTDRAG;
		 		                       decodeclick(&m);
                           }

					                      if(abxstate & (AB1|AB2))
                           {
                            oldstate=abxstate;

                            if(inside)
                            {
                             abxstate|=ABPUSH;
                            }
                            else
                            {
                             abxstate&=~ABPUSH;
                            }

                            if(oldstate!=abxstate) abxmouse(child,abxstate);
                           }

                           mousemove(&m);

                           break;

     case WM_RBUTTONDBLCLK:
                           wimp_checkfocus(hwnd);
      				                 abxstate|=ABRIGHTPUSH2;
                           abxmouse(child,abxstate);
                           XSetCapture(child,&hprev);  // has to be here because the click may launch another capture
                           if(abxstate & ABREPORTDOUBLECLICK)
                           {
                            m.bbits=(wimp_bbits)0x100;
                            decodeclick(&m);
                            abxstate&=~ABREPORTDOUBLECLICK;
                           }
                           abxauto(&m,abxstate);
						                     break;

       case WM_RBUTTONDOWN:
                           wimp_checkfocus(hwnd);
                           abxhandle=(wimp_w)hwnd;
						                     abxstate=ABRIGHTPUSH1;
                           abxsetstate(child);
                           abxmouse(child,abxstate);
                           XSetCapture(child,&hprev);   // has to be here because the click may launch another capture
                           if(abxstate & ABREPORTSINGLECLICK)
                           {
                            m.bbits=(wimp_bbits)0x1;
	 					                     decodeclick(&m);
                            abxstate&=~ABREPORTSINGLECLICK;
                           }
                           abxauto(&m,abxstate);
						                     break;

         case WM_RBUTTONUP://cprintf("r up");
                           oldstate=abxstate;
                           XReleaseCapture();    // has to be here because the click may launch another capture
          
                           if((oldstate & ABREPORTSINGLERELEASE) && inside)
                           {
                            m.bbits=(wimp_bbits)0x1;
	 					                     decodeclick(&m);
                            abxstate&=~ABREPORTSINGLERELEASE;
                           }

                           if((oldstate & ABREPORTDOUBLERELEASE) &&
                              (oldstate & AB2) && inside)
                           {
                            m.bbits=(wimp_bbits)0x100;
	 					                     decodeclick(&m);
                            abxstate&=~ABREPORTDOUBLERELEASE;
                           }

                           oldstate=abxstate;
 				                      abxstate&=~(AB1|AB2|ABRIGHT|ABPUSH|ABREPORTDRAG);
                           if(oldstate!=abxstate)
                           {
                            abxmouse(child,abxstate);
                           }
                           abxauto(&m,abxstate);
					                      break;
 }

 return(0);
 USE(wParam);
}



static os_error * abxmouseclose(wimp_w handle)
{
// cprintf("mouse close state %d handles %d",mfnstate,savemouse.w==handle);

 if(abxstate && abxhandle==handle)
 {
  XReleaseCapture();
  abxstate=0;
 }

 return(NULL);
}




// return 1 if we do something

static int abxsetcursor(MSG * msg)
{
 wimp_w        handle;
 int           hitcode;
 int           code;
 char          buffer[256];

 code=0;
 handle=(wimp_w)msg->hwnd;
 hitcode=LOWORD(msg->lParam);

 if(hitcode==HTCLIENT)
 {
  abxread(msg->hwnd,buffer,sizeof(buffer)/sizeof(TCHAR));

  if(getgivenitem(buffer,'P',buffer))
  {
   if(*buffer=='i')   // icon
   {
    changepointer(ICONPOINTER,NULL,buffer+1);
    code=1;
   }
   else
   if(*buffer=='s')  // system
   {


   }
   else
   if(*buffer=='a')  // sprite
   {


   }
  }
 }

 return(code);
}












static int abxpaint(HWND hwnd)
{
 PAINTSTRUCT paintstruct;
 HDC         hdc;
 RECT        rect;
 int         word;
 HFONT       font;

 if(GetUpdateRect(hwnd,NULL,FALSE))
 {
  word=GetWindowLong(hwnd,ABXFLAGS);
  font=(HFONT)GetWindowLongPtr(hwnd,ABXFONT);

  hdc=BeginPaint(hwnd,&paintstruct);
 
  XGetClientRect(hwnd,&rect);

  drawabx(hwnd,hdc,font,&rect,word,(sprite_area*)sprites);

  EndPaint(hwnd,&paintstruct);
 }

 return(0);
}


static int abxfocus(HWND hwnd,int gain)
{
 int word;

 word=GetWindowLong(hwnd,ABXFLAGS);
 if(gain) word|=USDFOCUS;
 else     word&=~USDFOCUS;
 SetWindowLong(hwnd,ABXFLAGS,word);

 InvalidateRect(hwnd,NULL,TRUE);
 UpdateWindow(hwnd);

 return(0);
}


static int abxsetfont(HWND hwnd,HFONT font)
{
// cprintf("abxset font %d",font);
 SetWindowLongPtr(hwnd,ABXFONT,(LONG_PTR)font);
 return(0);
}





os_error * abxgetslicewnd(HWND hwnd,char * string,int * max,int ctype,int * present)
{
 os_error * err;
 char       temp[1024];
 int        code;
 int        len;
 char     * s;
 char     * e;


 len=0;
 *string=0;
 if(present) *present=0;

 err=XDefWindowProc(hwnd,WM_GETTEXT,(WPARAM)(sizeof(temp)/sizeof(TCHAR)),(LPARAM)temp,(LRESULT*)&code);
 if(!err)
 {
  if(getitempointers(ctype,temp,&s,&e,NULL))
  {
   *e=0;
   s+=2;
   if(max) len=MIN((*max),e-s);
   else    len=sizeof(temp)/sizeof(TCHAR)-1;
   xstrncpy(string,s,len+1);
//   cprintf("string %s len %d",string,len);
   if(present) *present=1;
  }
 }

 if(max) *max=len;    // don't include terminating null

 return(err);
}


os_error * abxgetsliceicon(wimp_w w,int i,char * string,int * max,int ctype,int * present)
{
 os_error * err;
 HWND       hwnd;

          err=XGetDlgItem((HWND)w,i,&hwnd);
 if(!err) err=abxgetslicewnd(hwnd,string,max,ctype,present);

 return(err);
}





os_error * abxsetcolours(HWND hwnd,int fcol,int bcol)
{
 os_error * err;
 char       temp[1024];
 char       string[16];
 int        code;

 err=XDefWindowProc(hwnd,WM_GETTEXT,(WPARAM)(sizeof(temp)/sizeof(TCHAR)),(LPARAM)temp,(LRESULT*)&code);

 xsprintf(sizeof(string)/sizeof(TCHAR),NULL,string,"0x%x",fcol);
 abxslice('g',string,temp);
 xsprintf(sizeof(string)/sizeof(TCHAR),NULL,string,"0x%x",bcol);
 abxslice('k',string,temp);

 XDefWindowProc(hwnd,WM_SETTEXT,(WPARAM)0,(LPARAM)temp,(LRESULT*)&code);

 InvalidateRect(hwnd,NULL,TRUE);
 UpdateWindow(hwnd);

 return(err);
}



static int abxsettext(HWND hwnd,char * string)
{
 char temp[1024];
 char string2[1024];
 int  code;

 XDefWindowProc(hwnd,WM_GETTEXT,(WPARAM)(sizeof(temp)/sizeof(TCHAR)),(LPARAM)temp,(LRESULT*)&code);

 unquoteicontext(string2,string,sizeof(string2)/sizeof(TCHAR));
 abxslice('s',string2,temp);

 XDefWindowProc(hwnd,WM_SETTEXT,(WPARAM)0,(LPARAM)temp,(LRESULT*)&code);

 InvalidateRect(hwnd,NULL,TRUE);
 UpdateWindow(hwnd);

 return(1);
}



os_error * abxslicewnd(HWND hwnd,int ctype,char * string)
{
 os_error * err;
 char       temp[1024];
 char       string2[1024];
 int        code;

 err=XDefWindowProc(hwnd,WM_GETTEXT,(WPARAM)(sizeof(temp)/sizeof(TCHAR)),(LPARAM)temp,(LRESULT*)&code);
 if(!err)
 {
  unquoteicontext(string2,string,sizeof(string2)/sizeof(TCHAR));
  abxslice(ctype,string2,temp);

//  cprintf("write %s hwnd %x",temp,hwnd);

  err=XDefWindowProc(hwnd,WM_SETTEXT,(WPARAM)0,(LPARAM)temp,(LRESULT*)&code);
  if(!err)
  {
   InvalidateRect(hwnd,NULL,TRUE);
   UpdateWindow(hwnd);
  }
 }

 return(err);
}



os_error * abxsliceicon(wimp_w w,int i,int ctype,char * newparam)
{
 os_error * err;
 HWND       hwnd;

          err=XGetDlgItem((HWND)w,i,&hwnd);
 if(!err) err=abxslicewnd(hwnd,ctype,newparam);

 return(err);
}




os_error * abxwrite(HWND hwnd,char * string)
{
 os_error * err;
 int        code;

 err=XDefWindowProc(hwnd,WM_SETTEXT,(WPARAM)0,(LPARAM)string,(LRESULT*)&code);
 if(!err)
 {
  InvalidateRect(hwnd,NULL,TRUE);
  UpdateWindow(hwnd);
 }

 return(err);
}


os_error * abxread(HWND hwnd,char * string,int size)
{
 os_error * err;
 int        code;

 err=XDefWindowProc(hwnd,WM_GETTEXT,(WPARAM)size,(LPARAM)string,(LRESULT*)&code);

 return(err);
}


os_error * abxwriteicon(wimp_w window,int icon,char * string)
{
 os_error * err;
 HWND       hwnd;

 err=XGetDlgItem((HWND)window,icon,&hwnd);
 if(!err) err=abxwrite(hwnd,string);
 return(err);
}


os_error * abxreadicon(wimp_w window,int icon,char * string,int size)
{
 os_error * err;
 HWND       hwnd;

 err=XGetDlgItem((HWND)window,icon,&hwnd);
 if(!err) err=abxread(hwnd,string,size);
 return(err);
}




// handler for our special custom controls


LRESULT CALLBACK abxevent(HWND hwnd, UINT message, 
                           WPARAM wParam, LPARAM lParam)
{
 MSG msg;
 int code;
 int reverselogic;

 code=1;
 reverselogic=0;

 msg.hwnd=hwnd;
 msg.lParam=lParam;
 msg.message=message;
 msg.wParam=wParam;


// cprintf("abx mess %x",message);

 switch(message)
 {
              case WM_PAINT:
			                         code=abxpaint(hwnd);
			                         break;

#ifdef NEVER
              case WM_CREATE:
                             { int word;
                               word=GetWindowLongPtr(hwnd,GWL_EXSTYLE);
                               cprintf("create word=%x",word);
                             }
                             code=0;
                             break;
#endif

            case WM_KEYDOWN:
            case WM_SYSKEYDOWN:
            case WM_SYSDEADCHAR:
               case WM_CHAR:
               case WM_SYSCHAR:
                           // cprintf("abx key");
                            //code=keypress(&msg,0);
	                           break;

               case WM_COMMAND:
                               //cprintf("abx command");
                               break;

               case WM_DESTROY:
                            code=0;
	                           // PostQuitMessage(0);
	                           break;

               case WM_HELP:
                            //code=helpcommand(hwnd,message,wParam,lParam);
                            break;


  
      case WM_LBUTTONDBLCLK:
        case WM_LBUTTONDOWN:
          case WM_LBUTTONUP:
          case WM_MOUSEMOVE:
      case WM_RBUTTONDBLCLK:
        case WM_RBUTTONDOWN:
          case WM_RBUTTONUP:
                            code=abxmousefn(hwnd,message,wParam,lParam);
                            break;


           case WM_SETFOCUS:
          case WM_KILLFOCUS:
                            code=abxfocus(hwnd,message==WM_SETFOCUS);
                            break;

            case WM_SETFONT:
                            code=abxsetfont(hwnd,(HFONT)wParam);
                            break;

     case WM_CAPTURECHANGED:
                            // losing capture
                            code=abxcapture(hwnd,(HWND)lParam);
                            break;


         case WM_GETDLGCODE:
                            // DLGC_BUTTON - means can't be checked
                            // need DEFPUSHBUTTON to get returns which get passed as WM_COMMAND's to the parent
                            //return(DLGC_WANTMESSAGE);
                            return(DLGC_DEFPUSHBUTTON);  // prob need to return Radio for some
                            break;


              case BM_CLICK:
                            //cprintf("abx click");
                            code=0;
                            break;

             case BN_CLICKED:
                            //cprintf("abx clicked");
                            break;

              case BM_SETCHECK:
                            code=abxcheck(hwnd,wParam);
                            break;

												 		case	WM_ENABLE:
																													code=abxenable(hwnd,wParam);
																													break;

           case    WM_SETTEXT:
                            reverselogic=1;
                            code=abxsettext(hwnd,(char*)lParam);
                            break;


           case    WM_GETTEXT:
                            code=wParam;
                            abxgetslicewnd(hwnd,(char*)lParam,&code,'s',NULL);
                            reverselogic=(code!=0);
                            // we always handle this, but code can be zero or non-zero
                            // it is the string length
                            break;


          case WM_SETCURSOR:
                            reverselogic=1;
                            hourglasssetcursor(&code);
                            if(!code)
                            {
                             // let other code have a go
                             code=abxsetcursor(&msg);
                            }
                            break;


         case WM_ERASEBKGND:
                            reverselogic=1;
                            code=abxeraseback(hwnd,(HDC)wParam);
                            break;



                    default:
                            break;

 }

 // normal logic commands return 0 if they do something

 if((code==0)!=(reverselogic==0)) XDefWindowProc(hwnd,message,wParam,lParam,(LRESULT*)&code);

 return(code);
}







/*****************************************************************************/


int taskclasswindow(HWND hwnd)
{
 LONG_PTR xxx;

 XGetWindowLongPtr(hwnd,GWLP_WNDPROC,(LONG_PTR*)&xxx);
 return(xxx==(LONG_PTR)pollevent);
}




HWND getcontrolparentwindow(HWND hmsg)
{
 LONG_PTR xxx;

 while(1)
 {
  if(hmsg==NULL) break;
  XGetWindowLongPtr(hmsg,GWLP_WNDPROC,(LONG_PTR*)&xxx);
  if(xxx==(LONG_PTR)pollevent) break;
  XGetParent(hmsg,&hmsg);
  if(hmsg==NULL) break;
 }

 return(hmsg);
}


// return 1 for normal message processing
// the aim of all this messing about is to
// ensure that if we use WM_KEYDOWN it is not
// used to generate a WM_CHAR message


static int ourkey(MSG * msg,HWND parent)
{
 int code;
 LONG_PTR xxx;
 int reverselogic;


 code=1;
 reverselogic=0;


 if(scanwinmessage(WINSCANLEVEL0,msg,&code,&reverselogic))
 {
  if(msg->message==WM_KEYDOWN || msg->message==WM_SYSKEYDOWN)
  {
//   cprintf("ourkey");
  
   XGetWindowLongPtr(msg->hwnd,GWLP_WNDPROC,(LONG_PTR*)&xxx);
   if(xxx==(LONG_PTR)pollevent)
   {
    code=keypress(msg,0);
    if(code)     // we didn't use the key press
    {
//    cprintf("pass on key");

     if(wimp_isdialogue(parent) || !IsDialogMessage(parent,msg))
     {
      TranslateMessage(msg);
      keypressskip=1;
      if(pollflags & POLLUNICODE) DispatchMessageW(msg);
      else                        DispatchMessage(msg);

      keypressskip=0;
     }
    }
    return(0);
   }
  }


  if(parent==NULL || wimp_isdialogue(parent) || !IsDialogMessage(parent,msg))
  {
//   cprintf("trans %d",clock());
   TranslateMessage(msg);
   if(pollflags & POLLUNICODE) DispatchMessageW(msg);
   else                        DispatchMessage(msg);
  }
 
  return(1);
 }

 return(code);
}




void poll(void)
{
 os_error * err;
 MSG        msg;
 HWND       parent;
 int        code;

 /* leaves here */

 if(dopollidle)  
 {
  pollexit();

  if(pollflags & POLLUNICODE) code=PeekMessageW(&msg,NULL,0,0,PM_REMOVE);
  else                        code=PeekMessage(&msg,NULL,0,0,PM_REMOVE);

  if(code)
  {
   pollenter();
   if(msg.message==WM_QUIT)
   {
    terminate(&msg);
   }
   parent=getcontrolparentwindow(msg.hwnd);
   ourkey(&msg,parent);
  }
  else
  {
   pollenter();
   zero();
  }
 }
 else
 {
  pollexit();

  if(pollflags & POLLUNICODE) err=XGetMessageW(&msg,NULL,0,0,&code);
  else                        err=XGetMessage(&msg,NULL,0,0,&code);

  if(!err)
  {
   if(code)
   {
    pollenter();
    parent=getcontrolparentwindow(msg.hwnd);
    ourkey(&msg,parent);
   }
   else
   {
    pollenter();
    terminate(&msg);
   }
  }
  else
  {
   pollenter();
  }
 }
}



void pollquit(int code)
{
 XPostQuitMessage(code);
}



/*
void pollfake(wimp_eventstr * ep)
{

 USE(ep);
}

*/

/*
os_error * addwimpmessages(int * mess)
{

 return(NULL);

 USE(mess);
}


 */

// called by us (wimp.cpp) when we close a window
// chance to flush button handling code

os_error * pollclosewindow(wimp_w w)
{

 mouseclose(w);
 abxmouseclose(w);
 return(NULL);
}



os_error * polldeletewindow(wimp_w w)
{

 mouseclose(w);
 abxmouseclose(w);
 ddedelete(w);

 return(NULL);
}




// called by us (wimp.cpp) when we close a parent window
// chance to flush button handling code

os_error * pollclosechild(wimp_w child)
{

 mouseclose(child);
 abxmouseclose(child);
 wclosechild(child);

 return(NULL);
}



os_error * polldeletechild(wimp_w child)
{

 mouseclose(child);
 abxmouseclose(child);
 wdeletechild(child);
 ddedelete(child);

 return(NULL);
}



int pollsetflags(int orr,int bic)
{
 pollflags&=~bic;
 pollflags|=orr;

 return(pollflags);
}


os_error * pollinit(void)
{
 os_error * err;

 eventn=0;

 err=flex_alloc((flex_ptr)&eventstack,EVCHUNK);


 return(err);
}



/*

LRESULT CALLBACK ListProc (HWND hwnd, UINT message, 
                           WPARAM wParam, LPARAM lParam)
{
     if (message == WM_KEYDOWN && wParam == VK_RETURN)
          SendMessage (GetParent (hwnd), WM_COMMAND, 
                       MAKELONG (1, LBN_DBLCLK), (LPARAM) hwnd) ;
          
     return CallWindowProc (OldList, hwnd, message, wParam, lParam) ;
}



os_error * subclasskeys(wimp_w handle,int icon)
{

 OldList = (WNDPROC) SetWindowLong (hwndList, GWL_WNDPROC,
                                               (LPARAM) ListProc) ;
          

}


*/



static LRESULT CALLBACK subevent(HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam)
{
 int          hit;
 LRESULT      code=0;
 int          event;
 eventscanstr es;
 WNDPROC      oldproc;


 hit=0;
 oldproc=NULL;

// cprintf("subevent");

 starteventscan(&es);

 while((event=nextevent(&es,SUBCLASS,-1,(wimp_w)hwnd,-1))>=0)
 {
  oldproc=(WNDPROC)eventstack[event].message;
  eventstack[event].fn.sub(eventstack[event].icon,hwnd,message,wParam,lParam,&hit,&code);
  if(hit) break;
 }

 endeventscan(&es);

 if(!hit && oldproc)
 {
  code=CallWindowProc(oldproc,hwnd,message,wParam,lParam);
 }


 return(code);
}



static os_error * addsubclass(wimp_w handle,int icon,uservalue userhandle,subfn sfn)
{
 os_error * err;
 HWND       child;
 int        event;
 WNDPROC    oldproc;

 err=NULL;

 if(icon>=0) err=XGetDlgItem((HWND)handle,icon,&child);
 else        child=(HWND)handle;

 event=findblock(SUBCLASS,-1,(wimp_w)child,userhandle,(zerofn)sfn);
 if(event<0)
 {
  oldproc=(WNDPROC)GetWindowLongPtr(child,GWLP_WNDPROC);
  SetWindowLongPtr(child,GWLP_WNDPROC,(LONG_PTR)subevent);
 }
 else
 {
  oldproc=(WNDPROC)eventstack[event].message;
 }

 err=addblock(SUBCLASS,(uservalue)oldproc,(wimp_w)child,userhandle,(zerofn)sfn,NULL);

 return(err);
}





static os_error * remsubclass(wimp_w handle,int icon,uservalue userhandle,subfn sfn)
{
 os_error * err;
 int        event;
 HWND       child;
 WNDPROC    oldproc;

 err=NULL;

 if(icon>=0) err=XGetDlgItem((HWND)handle,icon,&child);
 else        child=(HWND)handle;


 event=findblock(SUBCLASS,-1,(wimp_w)child,userhandle,(zerofn)sfn);
 if(event>=0)
 {
  oldproc=(WNDPROC)eventstack[event].message;
  
  err=remblock(SUBCLASS,eventstack[event].message,(wimp_w)child,userhandle,(zerofn)sfn,NULL);
  
  event=findblock(SUBCLASS,-1,(wimp_w)child,userhandle,NULL);
  if(event<0)
  {
   SetWindowLongPtr(child,GWLP_WNDPROC,(LONG_PTR)oldproc);
  }
 }

 return(err);
}






static os_error * subkey(uservalue userhandle,HWND hwnd,UINT message,WPARAM wParam,LPARAM lParam,int * hit,LRESULT * code)
{
 MSG msg;
 HWND parent;

 parent=getcontrolparentwindow(hwnd);

 msg.hwnd=parent;
 msg.lParam=lParam;
 msg.message=message;
 msg.wParam=wParam;

// cprintf("subkey");

 switch(message)
 {
            case WM_KEYDOWN:
               case WM_CHAR:
                            *code=keypress(&msg,1);
                            if((*code)==0) *hit=1;
	                           break;
 

 }


 return(NULL);

 USE(userhandle);
}







os_error * addkeyeventicon(wimp_w handle,int icon)
{
 os_error * err;

 err=addsubclass(handle,icon,0,subkey);

 return(err);
}



os_error * remkeyeventicon(wimp_w handle,int icon)
{
 os_error * err;

 err=remsubclass(handle,icon,0,subkey);

 return(err);
}


